import logging

import pandas as pd
import pytest
from bluesky.preprocessors import run_wrapper
from numpy import linspace
from ophyd.sim import SynAxis

from ..plans.scans import centroid_scan
from ..utils import as_list
from .conftest import SynCamera

logger = logging.getLogger(__name__)

m1 = SynAxis(name="m1")
m2 = SynAxis(name="m2")
delay = SynAxis(name="delay")


def test_centroid_scan_returns_df(fresh_RE):
    # Simulated camera
    camera = SynCamera(m1, m2, delay, name="camera")
    # Create the plan

    def test_plan():
        delay_scan = (yield from centroid_scan(camera, delay, -1, 1, 3,
                                               detector_fields=[
                                                   'camera_centroid_x',
                                                   'camera_centroid_y']))
        # Check that we got dataframe
        assert isinstance(delay_scan, pd.DataFrame)
        # Check that all the entries were written to and none are nans
        assert True not in delay_scan.isnull().values
    # Run the plan
    fresh_RE(run_wrapper(test_plan()))


@pytest.mark.parametrize("steps", [3, 11])
@pytest.mark.parametrize("stop", [5, 10])
@pytest.mark.parametrize("start", [-5, -10])
def test_centroid_scan_returns_correct_df_index(fresh_RE, start, stop, steps):
    # Simulated camera
    camera = SynCamera(m1, m2, delay, name="camera")
    # Create the plan

    def test_plan():
        delay_scan = (yield from centroid_scan(camera, delay, start, stop,
                                               steps, detector_fields=[
                                                   'camera_centroid_x',
                                                   'camera_centroid_y']))
        # Check the number of steps
        assert delay_scan.shape[0] == steps
        # Check all the values are equal to linspace
        assert (delay_scan.index.values == linspace(start, stop, steps)).all()
    # Run the plan
    fresh_RE(run_wrapper(test_plan()))


@pytest.mark.parametrize("system_attrs", [([], []),
                                          (m1, m1.name),
                                          ([m1, m2], [m1.name, m2.name])])
@pytest.mark.parametrize("mot_fields", ['',
                                        'delay',
                                        ['delay', 'delay_setpoint']])
@pytest.mark.parametrize("det_fields", ['centroid_x',
                                        ['centroid_x',
                                         'centroid_y']])
def test_centroid_scan_returns_correct_columns(fresh_RE, mot_fields, det_fields,
                                               system_attrs):
    # Simulated camera
    camera = SynCamera(m1, m2, delay, name="camera")
    # Create the plan

    def test_plan():
        steps = 2               # Start and stop position
        delay_scan = (yield from centroid_scan(camera, delay, -1, 1, steps,
                                               detector_fields=det_fields,
                                               motor_fields=mot_fields,
                                               system=system_attrs[0],
                                               system_fields=system_attrs[1]))

        expected_length = (len(as_list(mot_fields)) or 1) \
            + len(as_list(det_fields)) \
            + len(as_list(system_attrs[1]))
        # Check the number of columns based in the inputs
        assert delay_scan.shape[1] == expected_length

        expected_columns = as_list(mot_fields or delay.name) \
            + as_list(system_attrs[1]) \
            + ["_".join([camera.name, fld]) for fld in as_list(
                det_fields)]
        # Check the columns are all what we expect them to be
        assert (delay_scan.columns == expected_columns).all()
    # Run the plan
    fresh_RE(run_wrapper(test_plan()))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import logging
import time
from collections import OrderedDict

import pytest
from ophyd.device import Device

from hxrsnd import pneumatic
from hxrsnd.pneumatic import PressureSwitch, ProportionalValve, SndPneumatics

from .conftest import fake_device, get_classes_in_module

logger = logging.getLogger(__name__)


@pytest.mark.parametrize("dev", get_classes_in_module(pneumatic, Device))
def test_devices_instantiate_and_run_ophyd_functions(dev):
    device = fake_device(dev)
    assert(isinstance(device.read(), OrderedDict))
    assert(isinstance(device.read_configuration(), OrderedDict))


def test_ProportionalValve_opens_and_closes_correctly():
    valve = fake_device(ProportionalValve)
    valve.open()
    time.sleep(.1)
    assert valve.position == "OPEN"
    assert valve.opened is True
    assert valve.closed is False
    valve.close()
    time.sleep(.1)
    assert valve.position == "CLOSED"
    assert valve.opened is False
    assert valve.closed is True


def test_PressureSwitch_reads_correctly():
    press = fake_device(PressureSwitch)
    press.pressure.sim_put(0)
    assert press.position == "GOOD"
    assert press.good is True
    assert press.bad is False
    press.pressure.sim_put(1)
    assert press.position == "BAD"
    assert press.good is False
    assert press.bad is True


def test_SndPneumatics_open_and_close_methods():
    vac = fake_device(SndPneumatics)
    for valve in vac._valves:
        valve.close()
    time.sleep(.1)
    vac.open()
    time.sleep(.1)
    for valve in vac._valves:
        assert valve.opened
    vac.close()
    time.sleep(.1)
    for valve in vac._valves:
        assert valve.closed



"""
Pneumatics for SnD
"""
import logging

from ophyd import Component as Cmp
from ophyd.signal import EpicsSignal, EpicsSignalRO

from .snddevice import SndDevice

logger = logging.getLogger(__name__)


class PneuBase(SndDevice):
    """
    Base class for the penumatics.
    """
    tab_component_names = True
    tab_whitelist = ['status']

    def status(self, status="", offset=0, print_status=True, newline=False):
        """
        Returns the status of the device.

        Parameters
        ----------
        status : str, optional
            The string to append the status to.

        offset : int, optional
            Amount to offset each line of the status.

        print_status : bool, optional
            Determines whether the string is printed or returned.

        newline : bool, optional
            Adds a new line to the end of the string.

        Returns
        -------
        status : str
            Status string.
        """
        status += "{0}{1:<16}|{2:^16}\n".format(" "*offset, self.desc+"",
                                                self.position)
        if newline:
            status += "\n"
        if print_status is True:
            logger.info(status)
        else:
            return status

    def __repr__(self):
        """
        Returns the status of the valve. Alias for status().

        Returns
        -------
        status : str
            Status string.
        """
        return self.status(print_status=False)


class ProportionalValve(PneuBase):
    """
    Class for the proportional pneumatic valves.

    Components
    ----------
    valve : EpicsSignal
        Valve control and readback pv.
    """
    tab_whitelist = ['close', 'closed', 'open', 'opened', 'position']
    valve = Cmp(EpicsSignal, ":VGP")

    def open(self):
        """
        Closes the valve.
        """
        if self.opened:
            logger.info("Valve currently open.")
        else:
            return self.valve.set(1, timeout=self.set_timeout)

    def close(self):
        """
        Closes the valve.
        """
        if self.closed:
            logger.info("Valve currently closed.")
        else:
            return self.valve.set(0, timeout=self.set_timeout)

    @property
    def position(self):
        """
        Returns the position of the valve.

        Returns
        -------
        position : str
            String saying the current position of the valve. Can be "OPEN" or
            "CLOSED".
        """
        if self.valve.get() == 1:
            return "OPEN"
        elif self.valve.get() == 0:
            return "CLOSED"
        else:
            return "UNKNOWN"

    @property
    def opened(self):
        """
        Returns if the valve is in the opened state.

        Returns
        -------
        opened : bool
            True if the valve is currently in the 'opened' position.
        """
        return (self.position == "OPEN")

    @property
    def closed(self):
        """
        Returns if the valve is in the closed state.

        Returns
        -------
        closed : bool
            True if the valve is currently in the 'closed' position.
        """
        return (self.position == "CLOSED")


class PressureSwitch(PneuBase):
    """
    Pressure switch.

    Components
    ----------
    pressure : EpicsSignalRO
        Pressure readbac signal.
    """
    tab_whitelist = ['bad', 'good', 'position']
    pressure = Cmp(EpicsSignalRO, ":GPS")

    @property
    def position(self):
        """
        Returns the position of the valve.

        Returns
        -------
        position : str
            String saying the current position of the valve. Can be "OPEN" or
            "CLOSED".
        """
        if self.pressure.get() == 0:
            return "GOOD"
        elif self.pressure.get() == 1:
            return "BAD"
        else:
            return "UNKNOWN"

    @property
    def good(self):
        """
        Returns if the pressure is in the 'good' state.

        Returns
        -------
        good : bool
            True if the pressure switch is in the 'good' state.
        """
        return (self.position == "GOOD")

    @property
    def bad(self):
        """
        Returns if the pressure is in the 'bad' state.

        Returns
        -------
        bad : bool
            True if the pressure switch is in the 'bad' state.
        """
        return (self.position == "BAD")


class SndPneumatics(SndDevice):
    """
    Class that contains the various pneumatic components of the system.

    Components
    ----------
    t1_valve : ProportionalValve
        Proportional valve on T1.

    t4_valve : ProportionalValve
        Proportional valve on T4.

    vac_valve : ProportionalValve
        Proportional valve on the overall system.

    t1_pressure : PressureSwitch
        Pressure switch on T1.

    t4_pressure : PressureSwitch
        Pressure switch on T4.

    vac_pressure : PressureSwitch
        Pressure switch on the overall system.
    """
    tab_whitelist = ['close', 'open', 'pressures', 'status', 'valves']

    t1_valve = Cmp(ProportionalValve, ":N2:T1", desc="T1 Valve")
    t4_valve = Cmp(ProportionalValve, ":N2:T4", desc="T4 Valve")
    vac_valve = Cmp(ProportionalValve, ":VAC", desc="Vacuum Valve")

    t1_pressure = Cmp(PressureSwitch, ":N2:T1", desc="T1 Pressure")
    t4_pressure = Cmp(PressureSwitch, ":N2:T4", desc="T4 Pressure")
    vac_pressure = Cmp(PressureSwitch, ":VAC", desc="Vacuum Pressure")

    def __init__(self, prefix, name=None, *args, **kwargs):
        super().__init__(prefix, name=name, *args, **kwargs)
        self._valves = [self.t1_valve, self.t4_valve, self.vac_valve]
        self._pressure_switches = [self.t1_pressure, self.t4_pressure,
                                   self.vac_pressure]

    def status(self, status="", offset=0, print_status=True, newline=False):
        """
        Returns the status of the vacuum system.

        Parameters
        ----------
        status : str, optional
            The string to append the status to.

        offset : int, optional
            Amount to offset each line of the status.

        print_status : bool, optional
            Determines whether the string is printed or returned.

        newline : bool, optional
            Adds a new line to the end of the string.

        Returns
        -------
        status : str
            Status string.
        """
        status += "\n{0}Pneumatics".format(" "*offset)
        status += "\n{0}{1}\n{0}{2:^16}|{3:^16}\n{0}{4}\n".format(
            " "*(offset+2), "-"*34, "Device", "State", "-"*34)
        for valve in self._valves:
            status += valve.status(offset=offset+2, print_status=False)
        for pressure in self._pressure_switches:
            status += pressure.status(offset=offset+2, print_status=False)

        if newline:
            status += "\n"
        if print_status is True:
            logger.info(status)
        else:
            return status

    def open(self):
        """
        Opens all the valves in the vacuum system.
        """
        logging.info("Opening valves in SnD system.")
        for valve in self._valves:
            valve.open()

    def close(self):
        """
        Opens all the valves in the vacuum system.
        """
        logging.info("Closing valves in SnD system.")
        for valve in self._valves:
            valve.close()

    @property
    def valves(self):
        """
        Prints the positions of all the valves in the system.
        """
        status = ""
        for valve in self._valves:
            status += valve.status(print_status=False)
        logger.info(status)

    @property
    def pressures(self):
        """
        Prints the pressures of all the pressure switches in the system.
        """
        status = ""
        for pressure in self._pressure_switches:
            status += pressure.status(print_status=False)
        logger.info(status)

    def __repr__(self):
        """
        Returns the status of the device. Alias for status().

        Returns
        -------
        status : str
            Status string.
        """
        return self.status(print_status=False)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import logging
import time
from collections import OrderedDict

import pytest
from ophyd.device import Device

from hxrsnd import attocube
from hxrsnd.attocube import EccBase, MotorDisabled, MotorError

from .conftest import fake_device, get_classes_in_module

logger = logging.getLogger(__name__)


@pytest.mark.parametrize("dev", get_classes_in_module(attocube, Device))
def test_attocube_devices_instantiate_and_run_ophyd_functions(dev):
    motor = fake_device(dev)
    assert(isinstance(motor.read(), OrderedDict))
    assert(isinstance(motor.read_configuration(), OrderedDict))


def test_EccBase_raises_MotorDisabled_if_moved_while_disabled():
    motor = fake_device(EccBase)
    motor.motor_error.sim_put(0)
    assert not motor.error
    motor.disable()
    assert not motor.enabled
    with pytest.raises(MotorDisabled):
        motor.move(10)


def test_EccBase_raises_MotorError_if_moved_while_faulted():
    motor = fake_device(EccBase)
    motor.motor_error.sim_put(0)
    assert not motor.error
    motor.enable()
    time.sleep(.5)
    assert motor.enabled
    motor.motor_error.sim_put(1)
    with pytest.raises(MotorError):
        motor.move(10)

# @pytest.mark.parametrize("position", [1])
# def test_EccBase_callable_moves_the_motor(position):
#     motor = EccBase("TEST")
#     motor.enable()
#     motor.limits = (0, 1)
#     assert motor.user_setpoint.get() != position
#     time.sleep(0.25)
#     motor(position, wait=False)
#     time.sleep(0.1)
#     assert motor.user_setpoint.get() == position



"""
Scans for HXRSnD
"""
import logging

import numpy as np
import pandas as pd
from bluesky import Msg
from bluesky.plan_stubs import abs_set, checkpoint, trigger_and_read
from bluesky.plans import scan
from bluesky.preprocessors import run_decorator, stub_wrapper
from bluesky.utils import short_uid as _short_uid
from pswalker.plans import measure_average
from pswalker.utils import field_prepend

from ..utils import as_list
from .preprocessors import return_to_start as _return_to_start

logger = logging.getLogger(__name__)


def linear_scan(motor, start, stop, num, use_diag=True, return_to_start=True,
                md=None, *args, **kwargs):
    """
    Linear scan of a motor without a detector.

    Performs a linear scan using the inputted motor, optionally using the
    diagnostics, and optionally moving the motor back to the original start
    position. This scan is different from the regular scan because it does not
    take a detector, and simply scans the motor.

    Parameters
    ----------
    motor : object
        any 'setable' object (motor, temp controller, etc.)

    start : float
        starting position of motor

    stop : float
        ending position of motor

    num : int
        number of steps

    use_diag : bool, optional
        Include the diagnostic motors in the scan.

    md : dict, optional
        metadata
    """
    # Save some metadata on this scan
    _md = {'motors': [motor.name],
           'num_points': num,
           'num_intervals': num - 1,
           'plan_args': {'num': num,
                         'motor': repr(motor),
                         'start': start,
                         'stop': stop},
           'plan_name': 'daq_scan',
           'plan_pattern': 'linspace',
           'plan_pattern_module': 'numpy',
           'plan_pattern_args': dict(start=start, stop=stop, num=num),
           'hints': {},
           }
    _md.update(md or {})

    # Build the list of steps
    steps = np.linspace(**_md['plan_pattern_args'])

    # Let's store this for now
    start = motor.position

    # Define the inner scan
    # @stage_decorator([motor])
    @run_decorator(md=_md)
    def inner_scan():

        for i, step in enumerate(steps):
            logger.info("\nStep {0}: Moving to {1}".format(i+1, step))
            grp = _short_uid('set')
            yield Msg('checkpoint')
            # Set wait to be false in set once the status object is implemented
            yield Msg('set', motor, step, group=grp, *args, **kwargs)
            yield Msg('wait', None, group=grp)
            yield from trigger_and_read([motor])

        if return_to_start:
            logger.info("\nScan complete. Moving back to starting position: {0}"
                        "\n".format(start))
            yield Msg('set', motor, start, group=grp, *args, **kwargs)
            yield Msg('wait', None, group=grp)

    return (yield from inner_scan())


def centroid_scan(detector, motor, start, stop, steps, average=None,
                  detector_fields=['stats2_centroid_x', 'stats2_centroid_y'],
                  motor_fields=None, system=None, system_fields=None,
                  filters=None, return_to_start=True, *args, **kwargs):
    """
    Performs a scan and returns the centroids of the inputted detector.

    The returned centroids can be absolute or relative to the initial value. The
    values are returned in a pandas DataFrame where the indices are the target
    motor positions.

    Parameters
    ----------
    detector : :class:`.BeamDetector`
        Detector from which to take the value measurements

    motor : :class:`.Motor`
        Main motor to perform the scan

    start : float
        Starting position of motor

    stop : float
        Ending position of motor

    steps : int
        Number of steps to take

    average : int, optional
        Number of averages to take for each measurement

    detector_fields : iterable, optional
        Fields of the detector to add to the returned dataframe

    motor_fields : iterable, optional
        Fields of the motor to add to the returned dataframe

    system : list, optional
        Extra devices to include in the datastream as we measure the average

    system_fields : list, optional
        Fields of the extra devices to add to the returned dataframe

    filters : dict, optional
        Key, callable pairs of event keys and single input functions that
        evaluate to True or False. For more infromation see
        :meth:`.apply_filters`

    return_to_start : bool, optional
        Move the scan motor back to its initial position after the scan

    Returns
    -------
    df : pd.DataFrame
        DataFrame containing the detector, motor, and system fields at every
        step of the scan.
    """
    average = average or 1
    system = as_list(system or [])
    all_devices = [motor] + system + [detector]

    # Ensure all fields are lists
    detector_fields = as_list(detector_fields)
    motor_fields = as_list(motor_fields or motor.name)
    system_fields = as_list(system_fields or [])

    # Get the full detector fields
    prep_det_fields = [field_prepend(fld, detector) for fld in detector_fields]
    # Put all the fields together into one list
    all_fields = motor_fields + system_fields + prep_det_fields

    # Build the dataframe with the centroids
    df = pd.DataFrame(columns=all_fields, index=np.linspace(start, stop, steps))

    # Create a basic measuring plan
    def per_step(detectors, motor, step):
        # Perform step
        yield from checkpoint()
        logger.debug("Measuring average at step {0} ...".format(step))
        yield from abs_set(motor, step, wait=True)
        # Measure the average
        reads = (yield from measure_average(all_devices, num=average,
                                            filters=filters, *args, **kwargs))
        # Fill the dataframe at this step with the centroid difference
        for fld in all_fields:
            df.loc[step, fld] = reads[fld]

    # Run the inner plan
    @_return_to_start(motor, perform=return_to_start)
    def inner():
        plan = scan([detector], motor, start, stop, steps, per_step=per_step)
        yield from stub_wrapper(plan)
    yield from inner()

    # Return the filled dataframe
    return df



"""
Small plans used in HXRSnD
"""

import logging
import math

from pswalker.plans import measure_average
from pswalker.utils import field_prepend

from ..utils import as_list

logger = logging.getLogger(__name__)

# Used to strip `run_wrapper` off of plan
# Should probably be added as bluesky PR


def block_run_control(msg):
    """
    Block open and close run messages
    """
    if msg.command in ['open_run', 'close_run']:
        return None
    return msg


def euclidean_distance(device, device_fields, targets, average=None,
                       filters=None):
    """
    Calculates the euclidean distance between the device_fields and targets.

    Parameters
    ----------
    device : :class:`.Device`
        Device from which to take the value measurements

    device_fields : iterable
        Fields of the device to measure

    targets : iterable
        Target value to calculate the distance from

    average : int, optional
        Number of averages to take for each measurement

    Returns
    -------
    distance : float
        The euclidean distance between the device fields and the targets.
    """
    average = average or 1
    # Turn things into lists
    device_fields = as_list(device_fields)
    targets = as_list(targets, len(device_fields))

    # Get the full detector fields
    prep_dev_fields = [field_prepend(fld, device) for fld in device_fields]

    # Make sure the number of device fields and targets is the same
    if len(device_fields) != len(targets):
        raise ValueError(
            "Number of device fields and targets must be the same."
            "Got {0} and {1}".format(len(device_fields), len(targets))
        )
    # Measure the average
    read = (yield from measure_average([device], num=average, filters=filters))
    # Get the squared differences between the centroids
    squared_differences = [(read[fld]-target)**2 for fld, target in zip(
        prep_dev_fields, targets)]
    # Combine into euclidean distance
    distance = math.sqrt(sum(squared_differences))
    return distance



from functools import wraps

from bluesky.plan_stubs import abs_set
from bluesky.plan_stubs import wait as plan_wait
from bluesky.utils import short_uid


def return_to_start(*devices, perform=True):
    """
    Decorator that will find the current positions of all the inputted devices,
    and them move them back to those positions after running the inner plan.
    """
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            # Get the initial positions of all the inputted devices
            initial_positions = {dev: dev.position for dev in devices}
            try:
                return (yield from func(*args, **kwargs))
            finally:
                # Start returning all the devices to their initial positions
                if perform:
                    group = short_uid('set')
                    for dev, pos in initial_positions.items():
                        yield from abs_set(dev, pos, group=group)
                    # Wait for all the moves to finish if they haven't already
                    yield from plan_wait(group=group)
        return wrapper
    return decorator



"""
Sequencer Class
"""
import logging

from ophyd import Component as Cmp
from ophyd.signal import EpicsSignal
from ophyd.status import wait as status_wait

from .snddevice import SndDevice

logger = logging.getLogger(__name__)


class SeqBase(SndDevice):
    """
    Base sequencer class.
    """
    tab_whitelist = ['start', 'stop']
    state_control = Cmp(EpicsSignal, ":PLYCTL")

    def start(self):
        """
        Start the sequencer.
        """
        status = self.state_control.set(1, timeout=self.timeout)
        status_wait(status)

    def stop(self):
        """
        Stop the sequencer.
        """
        status = self.state_control.set(0, timeout=self.timeout)
        status_wait(status)



"""
Tests for the bin ipython shell
"""
import logging
import sys

import pytest

from .conftest import requires_epics

logger = logging.getLogger(__name__)


@pytest.mark.timeout(60)
@requires_epics
@pytest.mark.skipif(sys.version_info < (3, 6), reason="requires python3.6")
def test_snd_devices_import_with_epics():
    import snd_devices  # noqa


# I couldn't quickly get this to pass with ophyd 1.2.0 (zlentz)
# @pytest.mark.timeout(60)
# @pytest.mark.skipif(sys.version_info < (3, 6), reason="requires python3.6")
# def test_snd_devices_import_no_epics():
#     import snd_devices



"""
Alignment plans for the HXRSnD
"""

import logging

import numpy as np
from bluesky.plan_stubs import abs_set, checkpoint
from bluesky.plans import list_scan
from bluesky.preprocessors import msg_mutator, subs_decorator
from lmfit.models import LorentzianModel
from pswalker.callbacks import LiveBuild
from pswalker.plans import measure_average

from ..exceptions import UndefinedBounds
from .plan_stubs import block_run_control

logger = logging.getLogger(__name__)


def maximize_lorentz(detector, motor, read_field, step_size=1,
                     bounds=None, average=None, filters=None,
                     position_field='user_readback', initial_guess=None):
    """
    Maximize a signal with a Lorentzian relationship to a motor

    The following plan does a linear step scan through the parameter space
    while collecting information to create a Lorentzian model. After the scan
    has completed, the created model will be queried to find the estimated
    motor position that will yield the absolute maximum of the Lorentz equation

    Parameters
    ----------
    detector : obj
        The object to be read during the plan

    motor : obj
        The object to be moved via the plan.

    read_field : str
        Field of detector to maximize

    nsteps : int, optional
        The step size used by the initial linear scan to create the Lorentzian
        model. A smaller step size will create a more accurate model, while a
        larger step size will increase the speed of the entire operation.

    bounds : tuple, optional
        The lower and higher limit of the search space. If no value is given
        the :attr:`.limits` property of the motor will be queried next. If this
        does not yield anything useful an exception will be raised

    average : int, optional
        The number of shots to average at every step of the scan. If left as
        None, no average will be used

    filters : dict, optional
        Filters used to drop shots from the analysis

    position_field : str, optional
        Motor field that will have the Lorentzian relationship with the given
        signal

    initial_guess : dict, optional
        Initial guess to the Lorentz model parameters of `sigma` `center`
        `amplitude`
    """
    average = average or 1
    # Define bounds
    if not bounds:
        try:
            bounds = motor.limits
            logger.debug("Bounds were not specified, the area "
                         "between %s and %s will searched",
                         bounds[0], bounds[1])
        except AttributeError as exc:
            raise UndefinedBounds("Bounds are not defined by motor {} or "
                                  "plan".format(motor.name)) from exc
    # Calculate steps
    steps = np.arange(bounds[0], bounds[1], step_size)
    # Include the last step even if this is smaller than the step_size
    steps = np.append(steps, bounds[1])
    # Create Lorentz fit and live model build
    fit = LorentzianModel(missing='drop')
    i_vars = {'x': position_field}
    model = LiveBuild(fit, read_field, i_vars, filters=filters,
                      average=average, init_guess=initial_guess)
    # update_every=len(steps)) # Set to fit only on last step

    # Create per_step plan
    def measure(detectors, motor, step):
        # Perform step
        logger.debug("Measuring average at step %s ...", step)
        yield from checkpoint()
        yield from abs_set(motor, step, wait=True)
        # Measure the average
        return (yield from measure_average([motor, detector],
                                           num=average,
                                           filters=filters))
    # Create linear scan
    plan = list_scan([detector], motor, steps, per_step=measure)

    @subs_decorator(model)
    def inner():
        # Run plan (stripping open/close run messages)
        yield from msg_mutator(plan, block_run_control)

        # Yield result of Lorentz model
        logger.debug(model.result.fit_report())
        max_position = model.result.values['center']

        # Check that the estimated position is reasonable
        if not bounds[0] < max_position < bounds[1]:
            raise ValueError("Predicted maximum position of {} is outside the "
                             "bounds {}".format(max_position, bounds))
        # Order move to maximum position
        logger.debug("Travelling to maximum of Lorentz at %s", max_position)
        yield from abs_set(motor, model.result.values['center'], wait=True)

    # Run the assembled plan
    yield from inner()
    # Return the fit
    return model


def rocking_curve(detector, motor, read_field, coarse_step, fine_step,
                  bounds=None, average=None, fine_space=5, initial_guess=None,
                  position_field='user_readback', show_plot=True):
    """
    Travel to the maxima of a bell curve

    The rocking curve scan is two repeated calls of :func:`.maximize_lorentz`.
    The first is a rough step scan which searches the area given by ``bounds``
    using ``coarse_step``, the idea is that this will populate the model enough
    such that we can do a more accurate scan of a smaller region of the search
    space. Once the rough scan is completed, the maxima of the fit is used as
    the center of the new fine scan that probes a region of space with a region
    twice as large as the ``fine_space`` parameter. After this, the motor is
    translated to the calculated maxima of the model

    Parameters
    ----------
    detector : obj
        The object to be read during the plan

    motor : obj
        The object to be moved via the plan.

    read_field : str
        Field of detector to maximize

    coarse_step : float
        Step size for the initial rough scan

    fine_step : float
        Step size for the fine scan

    bounds : tuple, optional
        Bounds for the original rough scan. If not provided, the soft limits of
        the motor are used

    average : int, optional
        Number of shots to average at each step

    fine_space : float, optional
        The amount to scan on either side of the rough scan result. Note that
        the rocking_curve will never tell the secondary scan to travel outside

    position_field : str, optional
        Motor field that will have the Lorentzian relationship with the given
        signal

    initial_guess : dict, optional
        Initial guess to the Lorentz model parameters of `sigma` `center`
        `amplitude`
        of the ``bounds``, so this region may be truncated.

    show_plot : bool, optional
        Create a plot displaying the progress of the `rocking_curve`
    """
    # Define bounds
    if not bounds:
        try:
            bounds = motor.limits
            logger.debug("Bounds were not specified, the area "
                         "between %s and %s will searched",
                         bounds[0], bounds[1])
        except AttributeError as exc:
            raise UndefinedBounds("Bounds are not defined by motor {} or "
                                  "plan".format(motor.name)) from exc
    if show_plot:
        # Create plot
        # subscribe first plot to rough_scan
        pass
    # Run the initial rough scan
    try:
        model = yield from maximize_lorentz(detector, motor, read_field,
                                            step_size=coarse_step,
                                            bounds=bounds, average=average,
                                            position_field=position_field,
                                            initial_guess=initial_guess)
    except ValueError as exc:
        raise ValueError("Unable to find a proper maximum value"
                         "during rough scan") from exc
    # Define new bounds
    center = model.result.values['center']
    bounds = (max(center - fine_space, bounds[0]),
              min(center + fine_space, bounds[1]))

    logger.info("Rough scan of region yielded maximum of %s, "
                "performing fine scan from %s to %s ...",
                center, bounds[0], bounds[1])

    if show_plot:
        # Highlight search space on first plot
        # Subscribe secondary plot
        pass

    # Run the initial rough scan
    try:
        fit = yield from maximize_lorentz(detector, motor, read_field,
                                          step_size=fine_step, bounds=bounds,
                                          average=average,
                                          position_field=position_field,
                                          initial_guess=model.result.values)
    except ValueError as exc:
        raise ValueError("Unable to find a proper maximum value"
                         "during fine scan") from exc

    if show_plot:
        # Draw final calculated max on plots
        pass

    return fit



"""
Tests for the scripts file in the top level directory
"""
import logging

import pytest

from .conftest import requires_epics

logger = logging.getLogger(__name__)


def scripts_import():
    import scripts  # noqa


@pytest.mark.timeout(60)
@requires_epics
def test_scripts_import_with_epics():
    scripts_import()

# I couldn't quickly get this to pass with ophyd 1.2.0 (zlentz)
# @pytest.mark.timeout(60)
# def test_scripts_import_no_epics():
#    scripts_import()



"""
Calibration of the delay macromotor
"""
import logging

import pandas as pd
from ophyd.utils import LimitError
from pswalker.plans import measure_average, walk_to_pixel

from ..utils import as_list
from .preprocessors import return_to_start as _return_to_start
from .scans import centroid_scan

logger = logging.getLogger(__name__)


def calibrate_motor(detector, detector_fields, motor, motor_fields,
                    calib_motors, calib_fields, start, stop, steps,
                    confirm_overwrite=True, *args, **kwargs):
    """Performs a calibration scan using the inputted detector, motor, and
    calibration motors, then configures the motor using the resulting
    calibration.

    Parameters
    ----------
    detector : :class:`.BeamDetector`
        Detector from which to take the value measurements

    detector_fields : iterable
        Fields of the detector to measure

    motor : :class:`.Motor`
        Main motor to perform the scan

    calib_motors : iterable, :class:`.Motor`
        Motor to calibrate each detector field with

    start : float
        Starting position of motor

    stop : float
        Ending position of motor

    steps : int
        Number of steps to take

    confirm_overwrite : bool, optional
        Prompt the user if this plan will overwrite an existing calibration

    Returns
    -------
    configs : tuple of dict
        Old configuration and the new configuration.
    """
    calib_motors = as_list(calib_motors)
    # Check for motor having a _calib field
    motor_config = motor.read_configuration()
    if motor_config['calib']['value'] and motor_config['motors']['value']:
        logger.warning("Running the calibration procedure will overwrite the "
                       "existing calibration.")
        # If a calibration is loaded, prompt the user for verification
        if confirm_overwrite:
            # Prompt the user about the move before making it
            try:
                response = input("\nConfirm Overwrite [y/n]: ")
            except Exception as e:
                logger.warning("Exception raised: {0}".format(e))
                response = "n"
            if response.lower() != "y":
                logger.info("\nCalibration cancelled.")
                return
            logger.debug("\nOverwrite confirmed.")

    # Perform the calibration scan
    df_calib, df_scan, scaling, start_pos = yield from calibration_scan(
        detector,
        detector_fields,
        motor,
        motor_fields,
        calib_motors,
        calib_fields,
        start, stop, steps,
        *args, **kwargs)

    # Load the calibration into the motor
    return motor.configure(calib=df_calib,
                           scan=df_scan,
                           motors=[motor]+calib_motors,
                           scale=scaling,
                           start=start_pos)


def calibration_scan(detector, detector_fields, motor, motor_fields,
                     calib_motors, calib_fields, start, stop, steps,
                     first_step=0.01, average=None, filters=None,
                     return_to_start=True, *args, **kwargs):
    """Performs a calibration scan for the main motor and returns a correction
    table for the calibration motors.

    This adds to ``calibration_scan`` by moving the motors to their original
    positions and and running the calibration calculation function. It returns
    the expected motor calibration given the results from the scan, as well as
    the dataframe from the scan with all the data.

    Parameters
    ----------
    detector : :class:`.BeamDetector`
        Detector from which to take the value measurements

    detector_fields : iterable
        Fields of the detector to measure

    motor : :class:`.Motor`
        Main motor to perform the scan

    calib_motors : iterable, :class:`.Motor`
        Motor to calibrate each detector field with

    start : float
        Starting position of motor

    stop : float
        Ending position of motor

    steps : int
        Number of steps to take

    first_step : float, optional
        First step to take on each calibration motor when performing the
        correction

    average : int, optional
        Number of averages to take for each measurement

    delay : float, optional
        Time to wait inbetween reads

    tolerance : float, optional
        Tolerance to use when applying the correction to detector field

    max_steps : int, optional
        Limit the number of steps the correction will take before exiting

    gradients : float, optional
        Assume an initial gradient for the relationship between detector value
        and calibration motor position

    return_to_start : bool, optional
        Move all the motors to their original positions after the scan has been
        completed

    Returns
    -------
    df_calibration : pd.DataFrame
        Dataframe containing the points to be used for the calibration by the
        calibration motors

    df_calibration_scan : pd.DataFrame
        DataFrame containing the positions of the detector,  motor, and
        calibration motor fields during the initial scan. The indices are the
        target motor positions.

    scaling : list
        List of the scaling values in units of motor egu / detector value used
        to calculate the calibration.

    start_positions : list
        List of starting positions used to perform the calibration calculation.
    """
    num = len(detector_fields)
    calib_motors = as_list(calib_motors)
    calib_fields = as_list(calib_fields or [m.name for m in calib_motors])
    motor_fields = as_list(motor_fields or motor.read().keys())
    if len(calib_motors) != num:
        raise ValueError("Must have same number of calibration motors as "
                         "detector fields.")
    if len(calib_fields) != num:
        raise ValueError("Must have same number of calibration fields as "
                         "detector fields.")

    @_return_to_start(motor, *calib_motors, perform=return_to_start)
    def inner():
        # Perform the main scan, reading the positions of all the devices
        logger.debug("Beginning calibration scan")
        df_scan = yield from calibration_centroid_scan(
            detector, motor, calib_motors,
            start, stop, steps,
            detector_fields=detector_fields,
            motor_fields=motor_fields,
            calib_fields=calib_fields,
            average=average,
            filters=filters)

        # Find the distance per detector value scaling and initial positions
        scaling, start_positions = yield from detector_scaling_walk(
            df_scan,
            detector,
            calib_motors,
            first_step=first_step,
            average=average,
            filters=filters,
            system=[motor],
            *args, **kwargs)

        # Build the calibration table
        df_calibration = build_calibration_df(df_scan, scaling, start_positions,
                                              detector)

        logger.debug("Completed calibration scan.")
        return df_calibration, df_scan, scaling, start_positions

    return (yield from inner())


def calibration_centroid_scan(detector, motor, calib_motors, start, stop, steps,
                              calib_fields=None, *args, **kwargs):
    """Performs a centroid scan producing a dataframe with the values of the
    detector, motor, and calibration motor fields.

    Parameters
    ----------
    detector : :class:`.BeamDetector`
        Detector from which to take the value measurements

    motor : :class:`.Motor`
        Main motor to perform the scan

    calib_motors : iterable
        Calibration motors

    start : float
        Starting position of motor

    stop : float
        Ending position of motor

    steps : int
        Number of steps to take

    average : int, optional
        Number of averages to take for each measurement

    detector_fields : iterable, optional
        Fields of the detector to add to the returned dataframe

    motor_fields : iterable, optional
        Fields of the motor to add to the returned dataframe

    calib_fields : list, optional
        Fields of the of the calibration motors to add to the returned dataframe

    Returns
    -------
    df : pd.DataFrame
        DataFrame containing the detector, motor, and calibration motor fields
        at every step of the scan.

    Raises
    ------
    ValueError
        If the inputted number of calibration motors does not have the same
        length as the number of calibration fields.
    """
    calib_fields = as_list(calib_fields or [m.name for m in calib_motors])

    # Make sure the same number of calibration fields as motors are passed
    if len(calib_motors) != len(calib_fields):
        raise ValueError("Must one calibration field for every calibration "
                         "motor, but got {0} fields for {1} motors.".format(
                             len(calib_fields), len(calib_motors)))

    # Perform the main scan, correctly passing the calibration parameters
    df = yield from centroid_scan(
        detector, motor, start, stop, steps, *args, system=calib_motors,
        system_fields=calib_fields, return_to_start=False, **kwargs
    )

    # Let's adjust the column names of the calib motors
    df.columns = [c+"_pre" if c in calib_fields else c for c in df.columns]
    return df


def detector_scaling_walk(df_scan, detector, calib_motors,
                          first_step=0.01, average=None, filters=None,
                          tolerance=1, delay=None, max_steps=5, system=None,
                          drop_missing=True, gradients=None, *args, **kwargs):
    """Performs a walk to to the detector value farthest from the current value
    using each of calibration motors, and then determines the motor to detector
    scaling

    Using the inputted scan dataframe, the plan loops through each detector
    field, then finds the value that is farthest from the current value, and
    then performs a walk_to_pixel to that value using the corresponding
    calibration motor. Since the final absolute position does not matter so long
    as it is recorded, if a RuntimeError or LimitError is raised, the plan will
    simply use the current motor position for the scaling calculation.

    Parameters
    ----------
    df_scan : pd.DataFrame
        Dataframe containing the results of a centroid scan performed using the
        detector, motor, and calibration motors.

    detector : :class:`.Detector`
        Detector from which to take the value measurements

    calib_motors : iterable, :class:`.Motor`
        Motor to calibrate each detector field with

    first_step : float, optional
        First step to take on each calibration motor when performing the
        correction

    average : int, optional
        Number of averages to take for each measurement

    delay : float, optional
        Time to wait inbetween reads

    tolerance : float, optional
        Tolerance to use when applying the correction to detector field

    max_steps : int, optional
        Limit the number of steps the correction will take before exiting

    drop_missing : bool, optional
        Choice to include events where event keys are missing

    gradients : float, optional
        Assume an initial gradient for the relationship between detector value
        and calibration motor position

    Returns
    -------
    scaling : list
        List of scales in the units of motor egu / detector value

    start_positions : list
        List of the initial positions of the motors before the walk
    """
    detector_fields = [col for col in df_scan.columns if detector.name in col]
    calib_fields = [col[:-4] for col in df_scan.columns if col.endswith("_pre")]
    if len(detector_fields) != len(calib_fields):
        raise ValueError("Must have same number of calibration fields as "
                         "detector fields, but got {0} and {1}.".format(
                             len(calib_fields), len(detector_fields)))

    # Perform all the initial necessities
    num = len(detector_fields)
    average = average or 1
    calib_motors = as_list(calib_motors)
    first_step = as_list(first_step, num, float)
    tolerance = as_list(tolerance, num)
    gradients = as_list(gradients, num)
    max_steps = as_list(max_steps, num)
    system = as_list(system or []) + calib_motors

    # Define the list that will hold the scaling
    scaling, start_positions = [], []

    # Now let's get the detector value to motor position conversion for each fld
    for i, (dfld, cfld, cmotor) in enumerate(zip(detector_fields,
                                                 calib_fields,
                                                 calib_motors)):
        # Get a list of devices without the cmotor we are inputting
        inp_system = list(system)
        inp_system.remove(cmotor)

        # Store the current motor and detector value and position
        reads = yield from measure_average([detector]+system,
                                           num=average,
                                           filters=filters)
        motor_start = reads[cfld]
        dfld_start = reads[dfld]

        # Get the farthest detector value we know we can move to from the
        # current position
        idx_max = abs(df_scan[dfld] - dfld_start).values.argmax()

        # Walk the cmotor to the first pre-correction detector entry
        try:
            logger.debug("Beginning walk to {0} on {1} using {2}".format(
                df_scan.iloc[idx_max][dfld], detector.name, cmotor.name))
            yield from walk_to_pixel(
                detector, cmotor, df_scan.iloc[idx_max][dfld], *args,
                filters=filters, gradient=gradients[i],
                target_fields=[dfld, cfld],
                first_step=first_step[i], tolerance=tolerance[i],
                system=inp_system, average=average, max_steps=max_steps[i],
                **kwargs
            )

        except RuntimeError:
            logger.warning("walk_to_pixel raised a RuntimeError for motor '{0}'"
                           ". Using its current position {1} for scale "
                           "calulation.".format(cmotor.desc, cmotor.position))
        except LimitError:
            logger.warning("walk_to_pixel tried to exceed the limits of motor "
                           "'{0}'. Using current position '{1}' for scale "
                           "calculation.".format(cmotor.desc, cmotor.position))

        # Get the positions and values we moved to
        reads = (yield from measure_average([detector]+system,
                                            num=average,
                                            filters=filters))
        motor_end = reads[cfld]
        dfld_end = reads[dfld]

        # Now lets find the conversion from signal value to motor distance
        scaling.append((motor_end - motor_start)/(dfld_end - dfld_start))
        # Add the starting position to the motor start list
        start_positions.append(motor_start)

    # Return the final scaling list
    return scaling, start_positions


def build_calibration_df(df_scan, scaling, start_positions, detector):
    """Takes the scan dataframe, scaling, and starting positions to build a
    calibration table for the calibration motors.

    The resulting dataframe will contain all the scan motor read fields as well
    as two columns for each calibration motor that has an absolute correction
    relative correction.

    Parameters
    ----------
    df_scan : pd.DataFrame
        Dataframe containing the results of a centroid scan performed using the
        detector, motor, and calibration motors.

    scaling : list
        List of scales in the units of motor egu / detector value

    start_positions : list
        List of the initial positions of the motors before the walk

    detector : :class:`.Detector`
        Detector from which to take the value measurements

    Returns
    -------
    df_calibration : pd.DataFrame
        Calibration dataframe that has all the scan motor fields and the
        corrections required of the calibration motors in both absolute and
        relative corrections.
    """
    # Get the fields being used in the scan df
    detector_fields = [col for col in df_scan.columns if detector.name in col]
    calib_fields = [col[:-4] for col in df_scan.columns if col.endswith("_pre")]
    motor_fields = [col for col in df_scan.columns
                    if col not in detector_fields and not col.endswith("_pre")]

    # Ensure these two are equal
    if len(detector_fields) != len(calib_fields):
        raise ValueError("Must have same number of calibration fields as "
                         "detector fields, but got {0} and {1}.".format(
                             len(calib_fields), len(detector_fields)))

    df_corrections = pd.DataFrame(index=df_scan.index)

    # Use the conversion to create an expected correction table
    for scale, start, cfld, dfld in zip(scaling, start_positions, calib_fields,
                                        detector_fields):
        # Absolute move to make to perform correction for this motor
        df_corrections[cfld+"_post"] = \
            start - (df_scan[dfld] - df_scan[dfld].iloc[0]) * scale

    # Put together the calibration table
    df_calibration = pd.concat([df_scan[motor_fields], df_corrections], axis=1)

    return df_calibration



"""
Script for abstract motor classes used in the SnD.
"""
import logging
import time
from collections import OrderedDict
from functools import reduce

import pandas as pd
from bluesky.preprocessors import run_wrapper
from ophyd.device import Component as Cmp
from ophyd.signal import Signal
from ophyd.utils import LimitError
from pcdsdevices.epics_motor import PCDSMotorBase
from pcdsdevices.interface import FltMvInterface

from .exceptions import InputError
from .plans.calibration import calibrate_motor
from .plans.preprocessors import return_to_start as _return_to_start
from .snddevice import SndDevice
from .utils import as_list

logger = logging.getLogger(__name__)


class SndMotor(FltMvInterface, SndDevice):
    """
    Base Sndmotor class that has methods common to all the various motors,
    even
    the non-EpicsMotor ones.
    """
    pass


class SndEpicsMotor(PCDSMotorBase, SndMotor):
    """
    SnD motor that inherits from EpicsMotor, therefore having all the relevant
    signals
    """
    direction_of_travel = Cmp(Signal)
    motor_spg = Cmp(Signal, value=2)


class SamMotor(SndMotor):
    offset_freeze_switch = Cmp(Signal)
    home_forward = Cmp(Signal)
    home_reverse = Cmp(Signal)

    def check_value(self, value, retries=5):
        """
        Check if the value is within the soft limits of the motor.

        Raises
        ------
        ValueError
        """
        if value is None:
            raise ValueError('Cannot write None to epics PVs')

        for i in range(retries):
            try:
                low_limit, high_limit = self.limits
                if not (low_limit <= value <= high_limit):
                    raise LimitError("Value {} outside of range: [{}, {}]"
                                     .format(value, low_limit, high_limit))
                return
            except TypeError:
                logger.warning("Failed to get limits, retrying...")
                if i == retries-1:
                    raise


# TODO: Add a centroid scanning method
# TODO: Add ability to save caibrations to disk
# TODO: Add ability to load calibrations from disk
# TODO: Add ability to display calibrations
# TODO: Add ability to change post-processing done to scan.
# TODO: Add ability to redo scaling on scan
class CalibMotor(SndDevice):
    """
    Provides the calibration macro methods.
    """
    tab_whitelist = ['calibrate', 'calibration', 'has_calib', 'use_calib']

    def __init__(self, prefix, name=None, calib_detector=None,
                 calib_motors=None, calib_fields=None, motor_fields=None,
                 *args, **kwargs):
        super().__init__(prefix, name=name, *args, **kwargs)
        self.calib_motors = calib_motors
        self.calib_fields = calib_fields
        self.motor_fields = motor_fields
        self.use_calib = False
        self._calib = OrderedDict()
        self.configure()

    def calibrate(self, start, stop, steps, average=100, confirm_overwrite=True,
                  detector=None, detector_fields=None, RE=None,
                  return_to_start=True, *args, **kwargs):
        """Performs a calibration scan for this motor and updates the
        configuration.

        Warning: This has not been commissioned.

        Parameters
        ----------
        start : float
            Starting position of motor

        stop : float
            Ending position of motor

        steps : int
            Number of steps to take

        average : int, optional
            Number of averages to take for each measurement

        confirm_overwrite : bool, optional
            Prompt the user if this plan will overwrite an existing calibration

        detector : :class:`.BeamDetector`, optional
            Detector from which to take the value measurements

        detector_fields : iterable, optional
            Fields of the detector to measure

        RE : RunEngine, optional
            Bluesky runengine instance to run the calibration plan.

        return_to_start : bool, optional
            Move all the motors to their original positions after the scan has been
            completed
        """
        # Remove this once the calibration routine has been tested
        logger.warning('Calibration functionality has not been commissioned.')

        # Use the inputted runengine or the parent's
        RE = RE or self.parent.RE
        detector = detector or self.calib_detector
        detector_fields = detector_fields or self.detector_fields

        # Make sure everything returns to the starting position when finished
        @_return_to_start(self, *self.calib_motors, perform=return_to_start)
        def inner():
            _, _ = yield from calibrate_motor(
                detector,
                detector_fields,
                self,
                self.motor_fields,
                self.calib_motors,
                self.calib_fields,
                start, stop, steps,
                average=average,
                confirm_overwrite=confirm_overwrite,
                return_to_start=False,
                *args, **kwargs)

        # Run the inner plan
        RE(run_wrapper(inner()))

    @property
    def calibration(self):
        """
        Returns the current calibration of the motor as a dictionary.

        Returns
        -------
        calibration : dict
            Dictionary containing calib, motors, scan, scale, and start
            calibration parameters.
        """
        if self.has_calib:
            config = self.read_configuration()
            # Grab the values of each of the calibration parameters
            calib = {fld: config[fld]['value']
                     for fld in ['calib', 'scan', 'scale', 'start']}
            # Make sure there are motors before we iterate through the list
            if config['motors']['value']:
                # If the motors have name attributes, just return those
                calib['motors'] = [mot.name if hasattr(mot, 'name') else mot
                                   for mot in config['motors']['value']]
            else:
                calib['motors'] = None
            return calib
        else:
            return None

    def configure(self, *, calib=None, motors=None, scan=None, scale=None,
                  start=None):
        """
        Configure the calib-motor's move parameters.

        Parameters
        ----------
        calib : DataFrame, optional
            Lookup table for move calibration. This represents set positions of
            auxiliary movers that should be chosen as we move our main macro.

        motors : list, optional
            List of calibration motors

        scan : pd.DataFrame, optional
            Dataframe of the centroid scan used to compute the correction table

        scale : list, optional
            List of scales in the units of motor egu / detector value

        start : list, optional
            List of the initial positions of the motors before the walk

        Returns
        -------
        configs : tuple of dict
            old_config, new_config
        """
        # Save prev for return statement
        prev_config = self.read_configuration()
        self._config_calib(calib, motors, scan, scale, start)

        # If we get a good calibration, change use_calib so we can use it
        if self.has_calib:
            self.use_calib = True

        # Return the previous and new configs
        return prev_config, self.read_configuration()

    def _config_calib(self, calib, motors, scan, scale, start):
        """
        Handle the calibration arguments, and update the config dictionary
        accordingly.

        Parameters
        ----------
        calib : DataFrame
            Lookup table for move calibration. This represents set positions of
            auxiliary movers that should be chosen as we move our main macro.

        motors : list
            List of calibration motors

        scan : pd.DataFrame
            Dataframe of the centroid scan used to compute the correction table

        scale : list
            List of scales in the units of motor egu / detector value

        start : list
            List of the initial positions of the motors before the walk
        """
        # Start with all the previous calibration parameters
        save_calib = OrderedDict(self._calib)
        motors = as_list(motors) or None

        # Add in the new parameters if they are not None or empty. If they are,
        # None or empty, check if they already exist as keys in the dict and
        # only add them if they do not.
        for key, value in {'calib': calib, 'motors': motors, 'scan': scan,
                           'scale': scale, 'start': start}.items():
            if value is not None or (value is None and key not in save_calib):
                save_calib[key] = {'value': value, 'timestamp': time.time()}

        # Now check all those changes, raising errors if needed
        self._check_calib(save_calib)
        # We made it through the check, therefore it is safe to use
        self._calib = save_calib

    def _check_calib(self, save_calib):
        """
        Internal method that checks the values passed in the calibration dict
        to make sure they are valid before actually saving it. It will raise
        errors or send warnings depending on the inputs.

        Raises
        ------
        InputError
            If a correction table is passed without motors or the correction
            table and motors have a mismatched number of columns and motors.

        TypeError
            If a correction table is passed that is not a dataframe.
        """
        # Let's get all the values we will update the calibration with
        calib = save_calib['calib']['value']
        motors = save_calib['motors']['value']
        scan = save_calib['scan']['value']
        scale = save_calib['scale']['value']
        start = save_calib['start']['value']

        # If no correction table is passed, then there isn't anything to check
        if calib is None:
            pass

        # We have a correction table but it isnt a Dataframe
        elif not isinstance(calib, pd.DataFrame):
            raise TypeError("Only Dataframes are supported for calibrations "
                            "tables at this time. Got a calibration of type "
                            "{0}.".format(type(calib)))

        # We have a correction table but no motors to correct with
        elif not motors:
            raise InputError("Inputted a correction table with no calibration "
                             "motors.")

        # We have a correction table and calibration motors, but they arent the
        # same length, so we cannot actually use it
        elif len(calib.columns) != len(motors):
            raise InputError("Mismatched calibration size and number of "
                             "motors. Got {0} columns for {1} motors.".format(
                                 len(calib.columns), len(motors)))

        # We have the correct correction table and motors but one of the of the
        # extra parameters were not passed, which is critical for corrected
        # motions but warn the user
        elif scan is None or not scale or not start:
            logger.warning("Inputted correction table and calibration motors "
                           "but not all configuration data. Some calibration "
                           "updating methods may not be functional!")

    def _calib_compensate(self, position, *args, **kwargs):
        """
        Perform the additional corrected motions if there is a valid calibration
        and the user has indicated that they want to perform them.

        Parameters
        ----------
        position
            Position to move to.

        Returns
        -------
        status : AndStatus
            Status objects of all the extra motions performed.
        """
        # Grab the current calibration
        calib = self._calib['calib']['value']
        motors = self._calib['motors']['value']
        status_list = []

        # Only perform the compensation if there is a valid calibration and we
        # want to use the calibration
        if not self.has_calib or not self.use_calib:
            return

        # Grab the two rows where the main motor position (column 0) is
        # closest to the inputted position
        top = calib.iloc[(calib.iloc[:, 0] - position).abs().argsort().iloc[:2]]
        first, second = top.iloc[0], top.iloc[1]

        # Get the slope between the lines between each of the motor values using
        # the first column as the x positions
        slopes = (second - first) / (second.iloc[0] - first.iloc[1])

        # Use the slope and the closest point to interpolate the motor positions
        # at the inputted position
        interpolated_row = slopes * (position-first[0]) + first

        # Move each calibration motor to the interpolated position
        for i, motor in enumerate(motors[1:]):
            status = motor.move(interpolated_row[i+1], *args, **kwargs)
            status_list.append(status)

        # Reduce all the status objects into one AndStatus object and return it
        return reduce(lambda x, y: x & y, status_list)

    @property
    def has_calib(self):
        """
        Indicator if there is a valid calibration that can be used to perform
        correction moves.

        Because the only requirements to perform a corrected move are the
        correction table and the calibration motors, they are the only
        calibration parameters that must be defines. These parameters must also
        pass ``_check_calib`` without raising any exceptions.

        Returns
        -------
        has_calib : bool
            True if there is a calibration that can be used for correction
            motion, False otherwise.
        """
        # Grab the current correction table and calibration motors
        calib = self._calib['calib']['value']
        _ = self._calib['motors']['value']

        # Return False if we dont have a correction table
        if calib is None:
            return False
        try:
            # If we make it through the check, we have a valid calibration
            self._check_calib(self._calib)
            return True
        except Exception:
            # An exception was raised, the config is somehow invalid
            return False

    @property
    def use_calib(self):
        """
        Returns whether the user indicated that corrected motions should be
        used.

        Returns
        -------
        use_calib : bool
            Internal indicator if the user wants to perform correction motions.
        """
        return self._use_calib

    @use_calib.setter
    def use_calib(self, indicator):
        """
        Setter for use_calib. Will warn the user if corrected motions are
        desired but there is no valid configuration to use.

        Parameters
        ----------
        indicator : bool
            Indicator for whether corrected motions should be performed or not.
        """
        self._use_calib = bool(indicator)
        if self._use_calib is True and not self.has_calib:
            logger.warning("use_calib is currently set to True but there is "
                           "no valid calibration to use")

    def read_configuration(self):
        return self._calib

    def describe_configuration(self):
        if not self._calib:
            return super().describe_configuration()
        if isinstance(self._calib['calib']['value'], pd.DataFrame):
            shape = self._calib['calib']['value'].shape
        else:
            shape = [len(self._calib)]
        return OrderedDict(**dict(calib=dict(source='calibrate', dtype='array',
                                             shape=shape)),
                           **super().describe_configuration())



"""
RTDs
"""
import logging

from .snddevice import SndDevice

logger = logging.getLogger(__name__)


class RTDBase(SndDevice):
    """
    Base class for the RTD.
    """
    pass


class OmegaRTD(RTDBase):
    """
    Class for the Omega RTD.
    """
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import logging

logger = logging.getLogger(__name__)

# Too hard to port to ophyd=1.2.0
# @pytest.mark.parametrize("dev", get_classes_in_module(sndsystem, Device))
# def test_devices_instantiate_and_run_ophyd_functions(dev):
#     device = fake_device(dev)
#     assert(isinstance(device.read(), OrderedDict))
#     assert(isinstance(device.read_configuration(), OrderedDict))



"""
Script to hold functions copied over from blutils that pertain to computing
the bragg angle of the HXRSnD crystals.
"""
import logging

import numpy as np

logger = logging.getLogger(__name__)

# Globals

_manual_energy = None

# Chemical Formula Aliases
alias = {'Air': 'N1.562O.42C.0003Ar.0094',
         'air': 'N1.562O.42C.0003Ar.0094',
         'YAG': 'Y3Al5O12',
         'yag': 'Y3Al5O12',
         'Sapphire': 'Al2O3',
         'sapphire': 'Al2O3',
         }

# Crystal Lattice parameters (a, b, c, alpha, beta, gamma)
# a,b,c in angstroms
# alpha, beta, gamma in degrees
lattice_parameters = {
    'Si': (5.4310205, 5.4310205, 5.4310205, 90, 90, 90),
}

# define units and constants
u = {
    'ang': 1e10,
}

# Trigonometric Functions


def sind(A):
    """
    Sin of an angle specified in degrees.

    Parameters
    ----------
    A : float
        Angle in degrees.

    Returns
    -------
    x : float
        Sin of the angle.
    """
    Arad = np.deg2rad(A)
    x = np.sin(Arad)
    return x


def cosd(A):
    """
    Cos of an angle specified in degrees.

    Parameters
    ----------
    A : float
        Angle in degrees.

    Returns
    -------
    x : float
        Cos of the angle.
    """
    Arad = np.deg2rad(A)
    x = np.cos(Arad)
    return x


def tand(A):
    """
    Tan of an angle specified in degrees.

    Parameters
    ----------
    A : float
        Angle in degrees.

    Returns
    -------
    x : float
        Tan of the angle.
    """
    Arad = np.deg2rad(A)
    x = np.tan(Arad)
    return x


def asind(x):
    """
    Calculates the arcsin in degrees.

    Parameters
    ----------
    x : float
        Value to calculate arcsin of.

    Returns
    -------
    A : float
        Arcsin of the value in degrees.
    """
    A = np.arcsin(x)
    A = np.rad2deg(A)
    return A


def acosd(x):
    """
    Calculates the arccos in degrees.

    Parameters
    ----------
    x : float
        Value to calculate arccos of.

    Returns
    -------
    A : float
        Arccos of the value in degrees.
    """
    A = np.arccos(x)
    A = np.rad2deg(A)
    return A


def atand(x):
    """
    Calculates the arctan in degrees.

    Parameters
    ----------
    x : float
        Value to calculate arctan of.

    Returns
    -------
    A : float
        Arctan of the value in degrees.
    """
    A = np.arctan(x)
    A = np.rad2deg(A)
    return A

# Frequency, wavelength and energy conversions


def lam(E, o=0):
    """
    Computes photon wavelength in m

    Parameters
    ----------
    E : float
        The input energy in eV or KeV

    o : float, optional
        Set o to 0 if working at sub-100 eV energies

    Returns
    -------
    lam : float
        Input energy converted to wavelength
    """
    if o:
        E = E
    else:
        E = eV(E)
    lam = (12398.4/E)/u['ang']
    return lam


def lam2E(l):
    """
    Computes photon energy in eV

    Parameters
    ----------
    l : float
        Photon wavelength in m

    Returns
    -------
    E : float
        Energy in eV
    """
    E = 12398.4/(l*u['ang'])
    return E


def lam2f(l):
    """
    Computes the photon frequency in Hz

    Parameters
    ----------
    l : float
        Photon wavelength in m

    Returns
    -------
    f : float
        Frequency in Hz
    """
    f = 299792458/l
    return f

# Higher level functions


def eV(E):
    """
    Returns photon energy in eV if specified in eV or KeV. Assumes that any
    value that is less than 100 is in KeV.

    Parameters
    ----------
    E : float
        The input energy to convert to eV

    Returns
    -------
    E : float
        Energy converted to eV from KeV
    """
    if E < 100:
        E *= 1000.0
    return float(E)


def check_id(ID):
    """
    Checks to see if you are using an alias. Returns the chemical formula

    Parameters
    ----------
    ID : str
        The desired ID

    Returns
    id : str
        The full ID name from the alias dictionary or just ID
    """
    try:
        return alias[ID]
    except KeyError:
        return ID


def get_e(energy=None, correct_ev=True):
    """
    Get working energy

    Parameters
    ----------
    energy : float or None, optional
        If energy passed in, return it, otherwise return manual_energy if set,
        or machine energy otherwise.

    correct_ev : bool, optional
        Convert to eV if True

    Returns
    -------
    en : float
        The desired working energy
    """
    en = None
    if energy is not None:
        en = energy
    elif _manual_energy is not None:
        en = _manual_energy
    if correct_ev:
        en = eV(en)
    return en


def d_space(ID, hkl):
    """
    Computes the d spacing (m) of the specified material and reflection

    Parameters
    ----------
    ID : str
        Chemical fomula : 'Si'

    hlk : tuple
        The reflection : (2,2,0)

    Returns
    -------
    d : float
        The d-spacing of the crystal using the inputted reflection.
    """
    ID = check_id(ID)
    h = hkl[0]
    k = hkl[1]
    l = hkl[2]

    lp = lattice_parameters[ID]
    a = lp[0]/u['ang']
    b = lp[1]/u['ang']
    c = lp[2]/u['ang']
    alpha = lp[3]
    beta = lp[4]
    gamma = lp[5]

    ca = cosd(alpha)
    cb = cosd(beta)
    cg = cosd(gamma)
    sa = sind(alpha)
    sb = sind(beta)
    sg = sind(gamma)

    invdsqr = 1 / (1.+2.*ca*cb*cg-ca**2.-cb**2.-cg**2.) * \
        (h**2.*sa**2./a**2. + k**2.*sb**2./b**2. + l**2.*sg**2./c**2. +
         2.*h*k*(ca*cb-cg)/a/b+2.*k*l*(cb*cg-ca)/b/c+2.*h*l*(ca*cg-cb)/a/c)

    d = invdsqr**-0.5
    return d


def bragg_angle(E=None, ID="Si", hkl=(2, 2, 0)):
    """
    Computes the Bragg angle (deg) of the specified material, reflection and
    photon energy.

    Parameters
    ----------
    E : float, optional
        Photon energy in eV or keV (default is LCLS value)

    ID : str, optional
        Chemical fomula : 'Si'

    hlk : tuple, optional
        The reflection : (2,2,0)

    Returns
    -------
    two_theta : float
        Expected bragg angle
    """
    ID = check_id(ID)
    E = get_e(energy=E, correct_ev=False)
    d = d_space(ID, hkl)
    two_theta = asind(lam(E)/2/d)
    return two_theta


def bragg_energy(theta, ID="Si", hkl=(2, 2, 0)):
    """
    Computes the photon energy that satisfies the Bragg condition of the
    specified material, reflection and theta angle.

    Parameters
    ----------
    theta : float, optional
        The scattering angle in degrees

    ID : str, optional
        Chemical fomula : 'Si'

    hlk : tuple, optional
        The reflection : (2,2,0)

    Returns
    -------
    E : float, optional
        Photon energy in eV
    """
    ID = check_id(ID)
    d = d_space(ID, hkl)
    l = 2*d*sind(theta)
    E = lam2E(l)
    return E


def snd_L(E1, E2, delay, gap=55):
    """
    Calculates the theta angles of the towers and the delay length based on the
    desired energy and delay.

    Parameters
    ----------
    E1 : float
        Energy of the delay branch in eV

    E2 : float
        Energy of the channel-cut branch in eV

    delay : float
        Delay of the system in picoseconds

    gap : float, optional
        Distance between the channel-cut crystals

    Returns
    -------
    theta_L : float
        The necessary angle of the delay branch in degrees.

    theta_cc : float
        The necessary angle of the channel-cut branch in degrees.

    L : float
        The necessary length of the delay crystals in mm.
    """
    cl = 0.3
    theta_L = bragg_angle('Si', (2, 2, 0), E1)
    theta_cc = bragg_angle('Si', (2, 2, 0), E2)
    # gap is the distance between the two faces of the channel cut crystal
    L = (delay*cl/2.+gap*(1-cosd(2*theta_cc))/sind(theta_cc))/(1-cosd(
        2*theta_L))
    logger.info("t1.L = t4.L = {} mm".format(L))
    logger.info("t1.tth = t4.tth = {} degree".format(2*theta_L))
    logger.info("t1.th1=t1.th2=t4.th1=t4.th2 = {} degree".format(theta_L))
    logger.info("t2.th=t3.th = {} degree".format(theta_cc))
    return theta_L, theta_cc, L


def snd_diag(E1, E2, delay, gap=55):
    """
    Calculates the positions of the middle diagnostics of the system based on
    the inputted energy and delay.

    Parameters
    ----------
    E1 : float
        Energy of the delay branch in eV

    E2 : float
        Energy of the channel-cut branch in eV

    delay : float
        Delay of the system in picoseconds

    gap : float, optional
        Distance between the channel-cut crystals

    Returns
    -------
    dd_x : float
        The necessary position of the middle delay diagnostic in mm

    dcc_x : float
        The necessary position of the middle channel-cut diagnostic in mm
    """
    cl = 0.3
    # speed of light
    theta_L = bragg_angle('Si', (2, 2, 0), E1)
    theta_cc = bragg_angle('Si', (2, 2, 0), E2)
    dcc_x = 2*cosd(theta_cc)*gap
    L = (delay*cl/2.+gap*(1-cosd(2*theta_cc))/sind(theta_cc))/(1-cosd(
        2*theta_L))
    dd_x = -L*sind(2*theta_L)
    logger.info("dd.x = {}".format(dd_x))
    logger.info("dcc.x = {}".format(dcc_x))
    return dd_x, dcc_x


def snd_delay(E1, E2, L, gap=55):
    """
    Calculates the delay of the system based on the inputted energies and the
    delay length.

    Parameters
    ----------
    E1 : float
        Energy of the delay branch in eV

    E2 : float
        Energy of the channel-cut branch in eV

    L : float
        Position of the delay crystals in mm

    Returns
    -------
    delay : float
        The delay of the system in picoseconds
    """
    cl = 0.3
    theta_L = bragg_angle('Si', (2, 2, 0), E1)
    theta_cc = bragg_angle('Si', (2, 2, 0), E2)
    delay = 2*(L*(1-cosd(2*theta_L)) - gap*(1-cosd(2*theta_cc))/sind(
        theta_cc))/cl
    return delay



"""
Script for small utility functions used in HXRSnD
"""
import inspect
import logging
from collections.abc import Iterable
from functools import wraps
from math import nan
from pathlib import Path

logger = logging.getLogger(__name__)


def absolute_submodule_path(submodule, cur_dir=inspect.stack()[0][1]):
    """
    Returns the absolute path of the inputted SnD submodule based on an inputted
    absolute path, or the absolute path of this file.

    Parameters
    ----------
    submodule : str or Path
        Desired submodule path.

    cur_dir : str or Path, optional
        Absolute path to use as a template for the full submodule path.

    Returns
    -------
    full_path : str
        Full string path to the inputted submodule.
    """
    dir_parts = Path(cur_dir).parts
    sub_parts = Path(submodule).parts
    base_path = Path(*dir_parts[:dir_parts.index(sub_parts[0])])
    if str(base_path) == ".":
        logger.warning("Could not match base path with desired submodule.")
    full_path = base_path / Path(submodule)
    return str(full_path)


def as_list(obj, length=None, tp=None, iter_to_list=True):
    """
    Force an argument to be a list, optionally of a given length, optionally
    with all elements cast to a given type if not None.

    Parameters
    ---------
    obj : Object
        The obj we want to convert to a list.

    length : int or None, optional
        Length of new list. Applies if the inputted obj is not an iterable and
        iter_to_list is false.

    tp : type, optional
        Type to cast the values inside the list as.

    iter_to_list : bool, optional
        Determines if we should cast an iterable (not str) obj as a list or to
        enclose it in one.

    Returns
    -------
    obj : list
        The object enclosed or cast as a list.
    """
    # If the obj is None, return empty list or fixed-length list of Nones
    if obj is None:
        if length is None:
            return []
        return [None] * length

    # If it is already a list do nothing
    elif isinstance(obj, list):
        pass

    # If it is an iterable (and not str), convert it to a list
    elif isiterable(obj) and iter_to_list:
        obj = list(obj)

    # Otherwise, just enclose in a list making it the inputted length
    else:
        try:
            obj = [obj] * length
        except TypeError:
            obj = [obj]

    # Cast to type; Let exceptions here bubble up to the top.
    if tp is not None:
        obj = [tp(o) for o in obj]
    return obj


def isiterable(obj):
    """
    Function that determines if an object is an iterable, not including
    str.

    Parameters
    ----------
    obj : object
        Object to test if it is an iterable.

    Returns
    -------
    bool : bool
        True if the obj is an iterable, False if not.
    """
    if isinstance(obj, str):
        return False
    else:
        return isinstance(obj, Iterable)


def _flatten(inp_iter):
    """
    Recursively iterate through values in nested iterables.

    Parameters
    ----------
    inp_iter : iterable
        The iterable to flatten.

    Returns
    -------
    value : object
        The contents of the iterable
    """
    for val in inp_iter:
        if isiterable(val):
            for ival in _flatten(val):
                yield ival
        else:
            yield val


def flatten(inp_iter):
    """
    Returns a flattened list of the inputted iterable.

    Parameters
    ----------
    inp_iter : iterable
        The iterable to flatten.

    Returns
    -------
    flattened_iter : list
        The contents of the iterable as a flat list
    """
    return list(_flatten(inp_iter))


def stop_on_keyboardinterrupt(func):
    """
    Decorator that runs the object's `stop` method if a keyboard interrupt is
    raised. This is meant to be used on ophyd device methods and expects the
    first argument to be `self`.
    """
    @wraps(func)
    def stop_dev_on_keyboardinterrupt(obj, *args, **kwargs):
        if hasattr(obj, "stop") and callable(obj.stop):
            try:
                return func(obj, *args, **kwargs)
            except KeyboardInterrupt:
                obj.stop()
                logger.info("Motor '{0}' stopped by keyboard interrupt".format(
                    obj.name))
        else:
            raise AttributeError("Object '{0}' needs a stop method to use the "
                                 "stop_on_keyboardinterrupt decorator.".format(
                                     obj))
    return stop_dev_on_keyboardinterrupt


def nan_if_no_parent(method):
    """
    Decorator that will return None if the object passed via self does not have
    a parent.
    """
    @wraps(method)
    def inner(obj, *args, **kwargs):
        if obj.parent:
            return method(obj, *args, **kwargs)
        else:
            return nan
    return inner



import logging
import math

import numpy as np
from bluesky.preprocessors import run_wrapper
from ophyd.sim import SynAxis

from ..plans.plan_stubs import euclidean_distance
from .conftest import SynCamera

logger = logging.getLogger(__name__)

m1 = SynAxis(name="m1")
m2 = SynAxis(name="m2")
delay = SynAxis(name="delay")


def test_euclidean_distance(fresh_RE):
    camera = SynCamera(m1, m2, delay, name="camera")
    # Define the distance plan that makes the assertion

    def test_plan():
        distance = yield from euclidean_distance(
            camera, ['centroid_x', 'centroid_y'], [1, 1])
        assert np.isclose(distance, math.sqrt(2))

    # Wrap the plan
    plan = run_wrapper(test_plan())
    # And now run it
    fresh_RE(plan)



"""
Script to hold the energy macromotors

All units of time are in picoseconds, units of length are in mm.
"""
import logging
from functools import reduce

from ophyd.device import Component as Cmp
from ophyd.signal import AttributeSignal
from ophyd.sim import NullStatus
from ophyd.status import wait as status_wait
from ophyd.utils import LimitError
from pcdsdevices.areadetector.detectors import PCDSAreaDetector
from pswalker.utils import field_prepend

from .bragg import bragg_angle, cosd, sind
from .exceptions import (BadN2Pressure, MotorDisabled, MotorFaulted,
                         MotorStopped)
from .sndmotor import CalibMotor, SndMotor
from .utils import flatten, nan_if_no_parent

logger = logging.getLogger(__name__)


_UNSET = object()


class MacroBase(SndMotor):
    """
    Base pseudo-motor class for the SnD macro-motions.
    """
    # Constants
    c = 0.299792458             # mm/ps
    gap = 55                    # m

    tab_component_names = True
    tab_whitelist = ['aligned', 'move', 'position', 'set', 'set_position',
                     'status', 'wait', 'c', 'gap']

    # Set add_prefix to be blank so cmp doesnt append the parent prefix
    readback = Cmp(AttributeSignal, "position", add_prefix='')

    def __init__(self, prefix, name=None, read_attrs=None, *args, **kwargs):
        read_attrs = read_attrs or ["readback"]
        super().__init__(prefix, name=name, read_attrs=read_attrs, *args,
                         **kwargs)
        self._status = NullStatus()
        self._use_diag = True
        self._verify_move_option = False

        # Make sure this is used
        if not self.parent:
            logger.warning("Macromotors must be instantiated with a parent "
                           "that has the SnD towers as components to function "
                           "properly.")
        else:
            self._delay_towers = [self.parent.t1, self.parent.t4]
            self._channelcut_towers = [self.parent.t2, self.parent.t3]

    @property
    def verify_move(self):
        """Verify that motion has completed after each move request."""
        return self._verify_move_option

    @verify_move.setter
    def verify_move(self, value):
        self._verify_move_option = bool(value)

    @property
    def use_diag(self):
        """Move the diagnostic motors to align with the beam."""
        return self._use_diag

    @use_diag.setter
    def use_diag(self, value):
        self._use_diag = bool(value)

    @property
    @nan_if_no_parent
    def position(self):
        """
        Returns the current position

        Returns
        -------
        energy : float
            Energy the channel cut line is set to in eV.
        """
        return (self.parent.t1.energy, self.parent.t2.energy,
                self._length_to_delay())

    def _verify_move(self, *args, **kwargs):
        """
        Prints a summary of the current positions and the proposed positions
        of the motors based on the inputs. It then prompts the user to confirm
        the move. To be overrided in subclasses.
        """
        pass

    def _check_towers_and_diagnostics(self, *args, **kwargs):
        """
        Checks the towers in the delay line and the channel cut line to make
        sure they can be moved. Depending on if E1, E2 or delay are entered,
        the delay line energy motors, channel cut line energy motors or the
        delay stages will be checked for faults, if they are enabled, if the
        requested energy requires moving beyond the limits and if the pressure
        in the tower N2 is good. To be overrided in subclasses.
        """
        pass

    def _move_towers_and_diagnostics(self, *args, **kwargs):
        """
        Moves all the tower and diagnostic motors according to the inputted
        energies and delay. If any of the inputted information is set to None
        then that component of the move is ignored. To be overrided in
        subclasses.
        """
        pass

    def _add_verify_header(self, string=""):
        """
        Adds the header that labels the motor desc, current position and propsed
        position.

        Parameters
        ----------
        string : str
            String to add a header to.

        Returns
        -------
        string : str
            String with the header added to it.
        """
        header = "\n{:^15}|{:^15}|{:^15}".format(
            "Motor", "Current", "Proposed")
        header += "\n" + "-"*len(header)
        return string + header

    def wait(self, status=None):
        """
        Waits for the status objects to complete the motions.

        Parameters
        ----------
        status : list or None, optional
            List of status objects to wait on. If None, will wait on the
            internal status list.
        """
        status = status or self._status
        logger.info("Waiting for the motors to finish moving...")
        status_wait(status)
        logger.info("Move completed.")

    def set(self, position, wait=True, verify_move=_UNSET, ret_status=True,
            use_diag=_UNSET):
        """
        Moves the macro-motor to the inputted position, optionally waiting for
        the motors to complete their moves.

        Parameters
        ----------
        position : float
            Position to move the macro-motor to.

        wait : bool, optional
            Wait for each motor to complete the motion before returning the
            console.

        verify_move : bool, optional
            Prints the current system state and a proposed system state and
            then prompts the user to accept the proposal before changing the
            system.

        ret_status : bool, optional
            Return the status object of the move.

        use_diag : bool, optional
            Move the daignostic motors to align with the beam.

        Returns
        -------
        status : list
            List of status objects for each motor that was involved in the move.
        """
        use_diag = use_diag if use_diag is not _UNSET else self.use_diag
        verify_move = (verify_move if verify_move is not _UNSET
                       else self._verify_move_option)
        return self.move(position, wait=wait, verify_move=verify_move,
                         use_diag=use_diag)

    def move(self, position, wait=True, verify_move=_UNSET, use_diag=_UNSET,
             *args, **kwargs):
        """
        Moves the macro-motor to the inputted position, optionally waiting for
        the motors to complete their moves. Alias for set().

        Parameters
        ----------
        position : float
            Position to move the macro-motor to.

        wait : bool, optional
            Wait for each motor to complete the motion before returning the
            console.

        verify_move : bool, optional
            Prints the current system state and a proposed system state and
            then prompts the user to accept the proposal before changing the
            system.

        use_diag : bool, optional
            Move the daignostic motors to align with the beam.

        Returns
        -------
        status : list
            List of status objects for each motor that was involved in the move.
        """
        if use_diag is _UNSET:
            use_diag = self.use_diag

        if verify_move is _UNSET:
            verify_move = self.verify_move

        # Check the towers and diagnostics
        diag_pos = self._check_towers_and_diagnostics(position,
                                                      use_diag=use_diag)

        # Prompt the user about the move before making it
        if verify_move and self._verify_move(position, use_diag=use_diag):
            return

        # Send the move commands to all the motors
        status_list = flatten(self._move_towers_and_diagnostics(
            position, diag_pos, use_diag=use_diag))

        # Aggregate the status objects
        status = reduce(lambda x, y: x & y, status_list)
        self._status = status

        # Wait for all the motors to finish moving
        if wait:
            self.wait(status)

        return status

    def mv(self, position, wait=True, verify_move=_UNSET, use_diag=_UNSET,
           *args, **kwargs):
        """
        Moves the macro parameters to the inputted positions. For energy, this
        moves the energies of both lines, energy1 moves just the delay line,
        energy2 moves just the channel cut line, and delay moves just the delay
        motors.

        mv() is different from move() by catching all the common exceptions that
        this motor can raise and just raises a logger warning. Therefore if
        building higher level functionality, do not use this method and use
        move() instead, otherwise none of the exceptions will propagate beyond
        this method.

        Parameters
        ----------
        position : float
            Position to move the motor to.

        wait : bool, optional
            Wait for each motor to complete the motion before returning the
            console.

        verify_move : bool, optional
            Prints the current system state and a proposed system state and
            then prompts the user to accept the proposal before changing the
            system.

        ret_status : bool, optional
            Return the status object of the move.

        use_diag : bool, optional
            Move the daignostic motors to align with the beam.

        Exceptions Caught
        -----------------
        LimitError
            Error raised when the inputted position is beyond the soft limits.

        MotorDisabled
            Error raised if the motor is disabled and move is requested.

        MotorFaulted
            Error raised if the motor is disabled and the move is requested.

        MotorStopped
            Error raised If the motor is stopped and a move is requested.

        BadN2Pressure
            Error raised if the pressure in the tower is bad.

        Returns
        -------
        status : list
            List of status objects for each motor that was involved in the move.
        """
        try:
            if use_diag is not _UNSET:
                old_use_diag = self.use_diag
                self.use_diag = use_diag

            if verify_move is not _UNSET:
                old_verify_move = self.verify_move
                self.verify_move = verify_move

            return super().mv(position, wait=wait, *args, **kwargs)
        # Catch all the common motor exceptions
        except LimitError:
            logger.warning("Requested move is outside the soft limits")
        except MotorDisabled:
            logger.warning("Cannot move '{0}' - a motor is currently disabled. "
                           "Try running 'motor.enable()'.".format(self.desc))
        except MotorFaulted:
            logger.warning("Cannot move '{0}' - a motor is currently faulted. "
                           "Try running 'motor.clear()'.".format(self.desc))
        except MotorStopped:
            logger.warning("Cannot move '{0}' - a motor is currently stopped. "
                           "Try running 'motor.state='Go''.".format(self.desc))
        except BadN2Pressure:
            logger.warning("Cannot move '{0}' - pressure in a tower is bad."
                           "".format(self.desc))
        finally:
            if use_diag is not _UNSET:
                self.use_diag = old_use_diag
            if verify_move is not _UNSET:
                self.verify_move = old_verify_move

    def set_position(self, *args, **kwargs):
        """
        Sets the current positions of the motors in the towers to be the
        calculated positions based on the inputted energies or delay. To be
        overrided in subclasses.
        """
        pass

    @property
    def aligned(self, *args, **kwargs):
        """
        Checks to see if the towers are in aligned in energy and delay. To be
        overrided in subclasses.
        """
        pass

    def _confirm_move(self, string):
        """
        Performs a basic confirmation of the move.

        Parameters
        ----------
        string : str
            Message to be printed to user.
        """
        logger.info(string)
        try:
            response = input("\nConfirm Move [y/n]: ")
        except Exception as e:
            logger.info("Exception raised: {0}".format(e))
            response = "n"

        if response.lower() != "y":
            logger.info("\nMove cancelled.")
            return True
        else:
            logger.debug("\nMove confirmed.")
            return False

    def status(self, status="", offset=0, print_status=True, newline=False):
        """
        Returns the status of the device.

        Parameters
        ----------
        status : str, optional
            The string to append the status to.

        offset : int, optional
            Amount to offset each line of the status.

        print_status : bool, optional
            Determines whether the string is printed or returned.
        """
        try:
            status += "\n{0}{1:<16} {2:^16}".format(
                " "*offset,
                self.desc+":",
                self.position)
        except TypeError:
            status += "\n{0}{1:<16} {2:^}".format(
                " "*offset,
                self.desc+":",
                str(self.position))

        if newline:
            status += "\n"
        if print_status:
            logger.info(status)
        else:
            return status


class DelayTowerMacro(MacroBase):
    """
    Class for the delay tower macros
    """

    def _delay_to_length(self, delay, theta1=None, theta2=None):
        """
        Converts the inputted delay to the lengths on the delay arm linear
        stages.

        Parameters
        ----------
        delay : float
            The desired delay in picoseconds.

        theta1 : float or None, optional
            Bragg angle the delay line is set to maximize.

        theta2 : float or None, optional
            Bragg angle the channel cut line is set to maximize.

        Returns
        -------
        length : float
            The distance between the delay crystal and the splitting or
            recombining crystal.
        """
        # Check if any other inputs were used
        theta1 = theta1 or self.parent.theta1
        theta2 = theta2 or self.parent.theta2

        # Length calculation
        length = ((delay*self.c/2 + self.gap*(1 - cosd(2*theta2)) / sind(theta2)) / (1 - cosd(2*theta1)))
        return length

    def _get_delay_diagnostic_position(self, E1=None, E2=None, delay=None):
        """
        Gets the position the delay diagnostic needs to move to based on the
        inputted energies and delay or the current bragg angles and current
        delay of the system.

        Parameters
        ----------
        E1 : float or None, optional
            Energy in eV to use for the delay line. Uses the current energy if
            None is inputted.

        E2 : float or None, optional
            Energy in eV to use for the channel cut line. Uses the current
            energy if None is inputted.

        delay : float or None, optional
            Delay in picoseconds to use for the calculation. Uses current delay
            if None is inputted.

        Returns
        -------
        position : float
            Position in mm the delay diagnostic should move to given the
            inputted parameters.
        """
        # Use current bragg angle
        if E1 is None:
            theta1 = self.parent.theta1
        else:
            theta1 = bragg_angle(E=E1)

        # Use current delay stage position if no delay is inputted
        if delay is None:
            length = self.parent.t1.length
        # Calculate the expected delay position if a delay is inputted
        else:
            if E2 is None:
                theta2 = self.parent.theta2
            else:
                theta2 = bragg_angle(E=E2)
            length = self._delay_to_length(delay, theta1=theta1, theta2=theta2)

        # Calculate position the diagnostic needs to move to
        position = -length*sind(2*theta1)
        return position


class DelayMacro(CalibMotor, DelayTowerMacro):
    """
    Macro-motor for the delay macro-motor.
    """
    # @property
    # def aligned(self, rtol=0, atol=0.001):
    #     """
    #     Checks to see if the towers are in aligned in energy and delay.

    #     Parameters
    #     ----------
    #     rtol : float, optional
    #         Relative tolerance to use when comparing value differences.

    #     atol : float, optional
    #         Absolute tolerance to use when comparing value differences.

    #     Returns
    #     -------
    #     is_aligned : bool
    #         True if the towers are aligned, False if not.
    #     """
    #     t1_delay = self._length_to_delay(self.parent.t1.length)
    #     t4_delay = self._length_to_delay(self.parent.t4.length)
    #     is_aligned = np.isclose(t1_delay, t4_delay, atol=atol, rtol=rtol)
    #     if not is_aligned:
    #         logger.warning("Delay mismatch between t1 and t4. t1: {0:.3f}ps, "
    #                        "t4: {1:.3f}ps".format(t1_delay, t4_delay))
    #     return is_aligned

    calib_detector = Cmp(PCDSAreaDetector, 'XCS:USR:O1000:01:', add_prefix=[])

    def __init__(self, prefix, name=None, *args, **kwargs):
        super().__init__(prefix, name=name, *args, **kwargs)
        if self.parent:
            self.motor_fields = ['readback']
            self.calib_motors = [self.parent.t1.chi1, self.parent.t1.y1]
            self.calib_fields = [field_prepend('user_readback', calib_motor)
                                 for calib_motor in self.calib_motors]
            self.detector_fields = ['stats2_centroid_x', 'stats2_centroid_y', ]

    def _length_to_delay(self, L=None, theta1=None, theta2=None):
        """
        Converts the inputted L of the delay stage, theta1 and theta2 to
        the expected delay of the system, or uses the current positions
        as inputs.

        Parameters
        ----------
        L : float or None, optional
            Position of the linear delay stage.

        theta1 : float or None, optional
            Bragg angle the delay line is set to maximize.

        theta2 : float or None, optional
            Bragg angle the channel cut line is set to maximize.

        Returns
        -------
        delay : float
            The delay of the system in picoseconds.
        """
        # Check if any other inputs were used
        L = L or self.parent.t1.length
        theta1 = theta1 or self.parent.theta1
        theta2 = theta2 or self.parent.theta2

        # Delay calculation
        delay = (2*(L*(1 - cosd(2*theta1)) - self.gap*(1 - cosd(2*theta2)) / sind(theta2))/self.c)
        return delay

    def _verify_move(self, delay, string="", use_header=True, confirm_move=True,
                     use_diag=_UNSET):
        """
        Prints a summary of the current positions and the proposed positions
        of the delay motors based on the inputs. It then prompts the user to
        confirm the move.

        Parameters
        ----------
        delay : float
            Desired delay of the system.

        string : str, optional
            Message to be printed as a prompt.

        use_header : bool, optional
            Adds a basic header to the message.

        confirm_move : bool, optional
            Prompts the user for confirmation.

        use_diag : bool, optional
            Add the diagnostic motor to the list of motors to verify.

        Returns
        -------
        allowed : bool
            True if the move is approved, False if the move is not.
        """
        use_diag = use_diag if use_diag is not _UNSET else self.use_diag
        # Add a header to the output
        if use_header:
            string += self._add_verify_header(string)

        # Convert to length and add the body of the string
        length = self._delay_to_length(delay)
        for tower in self._delay_towers:
            string += "\n{:<15}|{:^15.4f}|{:^15.4f}".format(
                tower.L.desc, tower.length, length)

        # Add the diagnostic move
        if use_diag:
            position_dd = self._get_delay_diagnostic_position(delay=delay)
            string += "\n{:<15}|{:^15.4f}|{:^15.4f}".format(
                self.parent.dd.x.desc, self.parent.dd.x.position, position_dd)

        # Prompt the user for a confirmation or return the string
        if confirm_move is True:
            return self._confirm_move(string)
        else:
            return string

    def _check_towers_and_diagnostics(self, delay, use_diag=_UNSET):
        """
        Checks the staus of the delay stages on the delay towers. Raises the
        basic motor errors if any of the motors are not ready to be moved.

        Parameters
        ----------
        delay : float
            The desired delay of the system.

        use_diag : bool, optional
            Check the position of the diagnostic motor.

        Raises
        ------
        LimitError
            Error raised when the inputted position is beyond the soft limits.

        MotorDisabled
            Error raised if the motor is disabled and move is requested.

        MotorFaulted
            Error raised if the motor is disabled and the move is requested.

        MotorStopped
            Error raised If the motor is stopped and a move is requested.

        BadN2Pressure
            Error raised if the pressure in the tower is bad.

        Returns
        -------
        position_dd : list
            Position to move the delay diagnostic to.
        """
        use_diag = use_diag if use_diag is not _UNSET else self.use_diag
        # Get the desired length for the delay stage
        length = self._delay_to_length(delay)

        # Check each of the delay towers
        for tower in self._delay_towers:
            tower.check_status(length=length)

        if use_diag:
            # Check the delay diagnostic position
            position_dd = self._get_delay_diagnostic_position(delay=delay)
            self.parent.dd.x.check_status(position_dd)
            return position_dd

    def _move_towers_and_diagnostics(self, delay, position_dd, use_diag=_UNSET):
        """
        Moves the delay stages and delay diagnostic according to the inputted
        delay and diagnostic position.

        Parameters
        ----------
        delay  : float
            Delay to set the system to.

        position_dd : float
            Position to move the delay diagnostic to.

        use_diag : bool, optional
            Move the daignostic motors to align with the beam.

        Returns
        -------
        status : list
            Nested list of status objects from each tower.
        """
        use_diag = use_diag if use_diag is not _UNSET else self.use_diag
        # Get the desired length for the delay stage
        length = self._delay_to_length(delay)

        # Move the delay stages
        status = [tower.set_length(length, wait=False, check_status=False)
                  for tower in self._delay_towers]
        # Log the delay change
        logger.debug("Setting delay to {0}.".format(delay))

        if use_diag:
            # Move the delay diagnostic to the inputted position
            status += [self.parent.dd.x.move(position_dd, wait=False)]

        # Perform the compensation
        if self.has_calib and self.use_calib:
            status.append(self._calib_compensate(delay))

        return status

    @property
    @nan_if_no_parent
    def position(self):
        """
        Returns the current energy of the channel cut line.

        Returns
        -------
        energy : float
            Energy the channel cut line is set to in eV.
        """
        return self._length_to_delay()

    def set_position(self, delay=None, print_set=True, use_diag=_UNSET,
                     verify_move=_UNSET):
        """
        Sets the current positions of the motors in the towers to be the
        calculated positions based on the inputted energies or delay.

        Parameters
        ----------
        delay : float or None, optional
            Delay to set the delay stages to.

        print_set : bool, optional
            Print a message to the console that the set has been made.

        verify_move : bool, optional
            Prints the current system state and a proposed system state and
            then prompts the user to accept the proposal before changing the
            system.
        """
        use_diag = use_diag if use_diag is not _UNSET else self.use_diag
        verify_move = (verify_move if verify_move is not _UNSET
                       else self._verify_move_option)

        # Prompt the user about the move before making it
        if verify_move and self._verify_move(delay, use_diag=use_diag):
            return

        length = self._delay_to_length(delay)

        # Set the position of each delay stage
        for tower in self._delay_towers:
            tower.L.set_position(length, print_set=False)

        # Diagnostic
        if use_diag:
            position_dd = self._get_delay_diagnostic_position(delay=delay)
            self.parent.dd.x.set_position(position_dd, print_set=False)

        if print_set:
            logger.info("Setting positions for delay to {0}.".format(delay))


class Energy1Macro(DelayTowerMacro):
    """
    Macro-motor for the energy 1 macro-motor.
    """
    # @property
    # def aligned(self, rtol=0, atol=0.001):
    #     """
    #     Checks to see if the towers are in aligned in energy and delay.

    #     Parameters
    #     ----------
    #     rtol : float, optional
    #         Relative tolerance to use when comparing value differences.

    #     atol : float, optional
    #         Absolute tolerance to use when comparing value differences.

    #     Returns
    #     -------
    #     is_aligned : bool
    #         True if the towers are aligned, False if not.
    #     """
    #     t1 = self.parent.t1
    #     t4 = self.parent.t4
    #     is_aligned = np.isclose(t1.energy, t4.energy, atol=atol, rtol=rtol)
    #     if not is_aligned:
    #         logger.warning("Energy mismatch between t1 and t4. t1: {0:.3f}  eV,"
    #                        " t4: {1:.3f} eV".format(t1.energy, t4.energy))
    #     return is_aligned

    def _length_to_delay(self, L=None, theta1=None, theta2=None):
        """
        Converts the inputted L of the delay stage, theta1 and theta2 to
        the expected delay of the system, or uses the current positions
        as inputs.

        Parameters
        ----------
        L : float or None, optional
            Position of the linear delay stage.

        theta1 : float or None, optional
            Bragg angle the delay line is set to maximize.

        theta2 : float or None, optional
            Bragg angle the channel cut line is set to maximize.

        Returns
        -------
        delay : float
            The delay of the system in picoseconds.
        """
        # Check if any other inputs were used
        L = L or self.parent.t1.length
        theta1 = theta1 or self.parent.theta1
        theta2 = theta2 or self.parent.theta2

        # Delay calculation
        delay = (2*(L*(1 - cosd(2*theta1)) - self.gap*(1 - cosd(2*theta2)) / sind(theta2))/self.c)
        return delay

    def _verify_move(self, E1, string="", use_header=True, confirm_move=True,
                     use_diag=_UNSET):
        """
        Prints a summary of the current positions and the proposed positions
        of the motors based on the inputs. It then prompts the user to confirm
        the move.

        Parameters
        ----------
        E1 : float
            Desired energy for the delay line.

        string : str, optional
            Message to be printed as a prompt.

        use_header : bool, optional
            Adds a basic header to the message.

        confirm_move : bool, optional
            Prompts the user for confirmation.

        use_diag : bool, optional
            Add the diagnostic motor to the list of motors to verify.

        Returns
        -------
        allowed : bool
            True if the move is approved, False if the move is not.
        """
        use_diag = use_diag if use_diag is not _UNSET else self.use_diag
        # Add a header to the output
        if use_header:
            string += self._add_verify_header(string)

        # Get move for each motor in the delay towers
        for tower in self._delay_towers:
            for motor, position in zip(tower._energy_motors,
                                       tower._get_move_positions(E1)):
                string += "\n{:<15}|{:^15.4f}|{:^15.4f}".format(
                    motor.desc, motor.position, position)

        if use_diag:
            position_dd = self._get_delay_diagnostic_position(E1)
            string += "\n{:<15}|{:^15.4f}|{:^15.4f}".format(
                self.parent.dd.x.desc, self.parent.dd.x.position, position_dd)

        # Prompt the user for a confirmation or return the string
        if confirm_move is True:
            return self._confirm_move(string)
        else:
            return string

    def _check_towers_and_diagnostics(self, E1, use_diag=_UNSET):
        """
        Checks the staus of the delay tower energy motors. Raises the basic
        motor errors if any of the motors are not ready to be moved

        Parameters
        ----------
        E1 : float
            Desired energy for the delay line.

        use_diag : bool, optional
            Check the position of the diagnostic motor.

        Raises
        ------
        LimitError
            Error raised when the inputted position is beyond the soft limits.

        MotorDisabled
            Error raised if the motor is disabled and move is requested.

        MotorFaulted
            Error raised if the motor is disabled and the move is requested.

        MotorStopped
            Error raised If the motor is stopped and a move is requested.

        BadN2Pressure
            Error raised if the pressure in the tower is bad.

        Returns
        -------
        position_dd : float or None
            Position to move the delay diagnostic to.
        """

        use_diag = use_diag if use_diag is not _UNSET else self.use_diag

        # Check each of the delay towers
        for tower in self._delay_towers:
            tower.check_status(energy=E1)

        # Check the delay diagnostic position
        if use_diag:
            position_dd = self._get_delay_diagnostic_position(E1=E1)
            self.parent.dd.x.check_status(position_dd)
            return position_dd

    def _move_towers_and_diagnostics(self, E1, position_dd, use_diag=_UNSET):
        """
        Moves the delay line energy motors and diagnostic to the inputted energy
        and diagnostic position.

        Parameters
        ----------
        E1 : float
            Energy to set the delay line to.

        position_dd : float
            Position to move the delay diagnostic to.

        use_diag : bool, optional
            Move the daignostic motors to align with the beam.

        Returns
        -------
        status : list
            Nested list of status objects from each tower.
        """
        use_diag = use_diag if use_diag is not _UNSET else self.use_diag

        # Move the towers to the specified energy
        status = [tower.set_energy(E1, wait=False, check_status=False) for
                  tower in self._delay_towers]
        # Log the energy change
        logger.debug("Setting E1 to {0}.".format(E1))

        # Move the delay diagnostic to the inputted position
        if use_diag:
            status += [self.parent.dd.x.move(position_dd, wait=False)]

        return status

    def _get_delay_diagnostic_position(self, E1=None):
        """
        Gets the position the delay diagnostic needs to move to based on the
        inputted energies and delay or the current bragg angles and current
        delay of the system.

        Parameters
        ----------
        E1 : float or None, optional
            Energy in eV to use for the delay line. Uses the current energy if
            None is inputted.

        Returns
        -------
        position : float
            Position in mm the delay diagnostic should move to given the
            inputted parameters.
        """
        return super()._get_delay_diagnostic_position(E1=E1)

    @property
    @nan_if_no_parent
    def position(self):
        """
        Returns the current energy of the delay line.

        Returns
        -------
        energy : float
            Energy the channel cut line is set to in eV.
        """
        return self.parent.t1.energy

    def set_position(self, E1=None, print_set=True, verify_move=_UNSET,
                     use_diag=_UNSET):
        """
        Sets the current positions of the motors in the towers to be the
        calculated positions based on the inputted energies or delay.

        Parameters
        ----------
        E1 : float or None, optional
            Energy to set the delay line to.

        print_set : bool, optional
            Print a message to the console that the set has been made.

        verify_move : bool, optional
            Prints the current system state and a proposed system state and
            then prompts the user to accept the proposal before changing the
            system.

        use_diag : bool, optional
            Move the daignostic motors to align with the beam.
        """
        use_diag = use_diag if use_diag is not _UNSET else self.use_diag
        verify_move = (verify_move if verify_move is not _UNSET
                       else self._verify_move_option)

        # Prompt the user about the move before making it
        if verify_move and self._verify_move(E1, use_diag=use_diag):
            return

        # Set position of each E1 motor in each delay tower
        for tower in self._delay_towers:
            for motor, pos in zip(tower._energy_motors,
                                  tower._get_move_positions(E1)):
                motor.set_position(pos, print_set=False)

        # Set the diagnostic
        if use_diag:
            position_dd = self._get_delay_diagnostic_position(E1=E1)
            self.parent.dd.x.set_position(position_dd, print_set=False)

        # Log the set
        if print_set is True:
            logger.info("Setting positions for E1 to {0}.".format(E1))


class Energy1CCMacro(Energy1Macro):
    """
    Macro-motor for the energy 1 channel cut macro-motor.
    """

    def _verify_move(self, E1, string="", use_header=True, confirm_move=True,
                     use_diag=_UNSET):
        """
        Prints a summary of the current positions and the proposed positions
        of the motors based on the inputs. It then prompts the user to confirm
        the move.

        Parameters
        ----------
        E1 : float
            Desired energy for the delay line.

        string : str, optional
            Message to be printed as a prompt.

        use_header : bool, optional
            Adds a basic header to the message.

        confirm_move : bool, optional
            Prompts the user for confirmation.

        use_diag : bool, optional
            Add the diagnostic motor to the list of motors to verify.

        Returns
        -------
        allowed : bool
            True if the move is approved, False if the move is not.
        """
        use_diag = use_diag if use_diag is not _UNSET else self.use_diag
        # Add a header to the output
        if use_header:
            string += self._add_verify_header(string)

        # Get move for each motor in the delay towers
        for tower in self._delay_towers:
            string += "\n{:<15}|{:^15.4f}|{:^15.4f}".format(
                tower.tth.desc, tower.tth.position, 2*bragg_angle(E1))

        if use_diag:
            position_dd = self._get_delay_diagnostic_position(E1)
            string += "\n{:<15}|{:^15.4f}|{:^15.4f}".format(
                self.parent.dd.x.desc, self.parent.dd.x.position, position_dd)

        # Prompt the user for a confirmation or return the string
        if confirm_move is True:
            return self._confirm_move(string)
        else:
            return string

    def _check_towers_and_diagnostics(self, E1, use_diag=_UNSET):
        """
        Checks the staus of the delay tower energy motors. Raises the basic
        motor errors if any of the motors are not ready to be moved

        Parameters
        ----------
        E1 : float
            Desired energy for the delay line.

        use_diag : bool, optional
            Check the position of the diagnostic motor.

        Raises
        ------
        LimitError
            Error raised when the inputted position is beyond the soft limits.

        MotorDisabled
            Error raised if the motor is disabled and move is requested.

        MotorFaulted
            Error raised if the motor is disabled and the move is requested.

        MotorStopped
            Error raised If the motor is stopped and a move is requested.

        BadN2Pressure
            Error raised if the pressure in the tower is bad.

        Returns
        -------
        position_dd : float or None
            Position to move the delay diagnostic to.
        """
        use_diag = use_diag if use_diag is not _UNSET else self.use_diag
        # Check each of the delay towers
        for tower in self._delay_towers:
            tower.tth.check_status(2*bragg_angle(E1))

        # Check the delay diagnostic position
        if use_diag:
            position_dd = self._get_delay_diagnostic_position(E1=E1)
            self.parent.dd.x.check_status(position_dd)
            return position_dd

    def _move_towers_and_diagnostics(self, E1, position_dd, use_diag=_UNSET):
        """
        Moves the delay line energy motors and diagnostic to the inputted energy
        and diagnostic position.

        Parameters
        ----------
        E1 : float
            Energy to set the delay line to.

        position_dd : float
            Position to move the delay diagnostic to.

        use_diag : bool, optional
            Move the daignostic motors to align with the beam.

        Returns
        -------
        status : list
            Nested list of status objects from each tower.
        """
        use_diag = use_diag if use_diag is not _UNSET else self.use_diag
        # Move the towers to the specified energy
        status = [tower.tth.move(2*bragg_angle(E1), wait=False,
                                 check_status=False)
                  for tower in self._delay_towers]
        # Log the energy change
        logger.debug("Setting E1_cc to {0}.".format(E1))

        # Move the delay diagnostic to the inputted position
        if use_diag:
            status += [self.parent.dd.x.move(position_dd, wait=False)]

        return status

    def set_position(self, E1=None, print_set=True, verify_move=_UNSET,
                     use_diag=_UNSET):
        """
        Sets the current positions of the motors in the towers to be the
        calculated positions based on the inputted energies or delay.

        Parameters
        ----------
        E1 : float or None, optional
            Energy to set the delay line to.

        print_set : bool, optional
            Print a message to the console that the set has been made.

        verify_move : bool, optional
            Prints the current system state and a proposed system state and
            then prompts the user to accept the proposal before changing the
            system.

        use_diag : bool, optional
            Move the daignostic motors to align with the beam.
        """
        verify_move = (verify_move if verify_move is not _UNSET
                       else self._verify_move_option)
        use_diag = use_diag if use_diag is not _UNSET else self.use_diag
        # Prompt the user about the move before making it
        if verify_move and self._verify_move(E1, use_diag=use_diag):
            return

        # Set position of each E1 motor in each delay tower
        for tower in self._delay_towers:
            tower.tth.set_position(2*bragg_angle(E1), print_set=False)

        # Set the diagnostic
        if use_diag:
            position_dd = self._get_delay_diagnostic_position(E1=E1)
            self.parent.dd.x.set_position(position_dd, print_set=False)

        # Log the set
        if print_set is True:
            logger.info("Setting positions for E1 to {0}.".format(E1))


class Energy2Macro(MacroBase):
    """
    Macro-motor for the energy 2 macro-motor.
    """

    def _get_channelcut_diagnostic_position(self, E2=None):
        """
        Gets the position the channel cut diagnostic needs to move to based on
        the inputted energy or the current energy of the channel cut line.

        Parameters
        ----------
        E2 : float or None, optional
            Energy in eV to use for the channel cut line. Uses the current
            energy if None is inputted.

        Returns
        -------
        position : float
            Position in mm the delay diagnostic should move to given the
            inputted parameters.
        """
        # Use the current theta2 of the system or calc based on inputted energy
        if E2 is None:
            theta2 = self.parent.theta2
        else:
            theta2 = bragg_angle(E=E2)

        # Calculate position the diagnostic needs to move to
        position = 2*cosd(theta2)*self.gap
        return position

    # @property
    # def aligned(self, rtol=0, atol=0.001):
    #     """
    #     Checks to see if the towers are in aligned in energy and delay.

    #     Parameters
    #     ----------
    #     rtol : float, optional
    #         Relative tolerance to use when comparing value differences.

    #     atol : float, optional
    #         Absolute tolerance to use when comparing value differences.

    #     Returns
    #     -------
    #     is_aligned : bool
    #         True if the towers are aligned, False if not.
    #     """
    #     t2 = self.parent.t2
    #     t3 = self.parent.t3
    #     is_aligned = np.isclose(t2.energy, t3.energy, atol=atol, rtol=rtol)
    #     if not is_aligned:
    #         logger.warning("Energy mismatch between t2 and t3. t1: {0:.3f} eV, "
    #                        "t4: {1:.3f} eV".format(t2.energy, t3.energy))
    #     return is_aligned

    def _verify_move(self, E2, string="", use_header=True, confirm_move=True,
                     use_diag=_UNSET):
        """
        Prints a summary of the current positions and the proposed positions
        of the motors based on the inputs. It then prompts the user to confirm
        the move.

        Parameters
        ----------
        E2 : float
            Desired energy for the channel cut line.

        string : str, optional
            Message to be printed as a prompt.

        use_header : bool, optional
            Adds a basic header to the message.

        confirm_move : bool, optional
            Prompts the user for confirmation.

        use_diag : bool, optional
            Add the diagnostic motor to the list of motors to verify.

        Returns
        -------
        allowed : bool
            True if the move is approved, False if the move is not.
        """
        use_diag = use_diag if use_diag is not _UNSET else self.use_diag

        # Add a header to the output
        if use_header:
            string += self._add_verify_header(string)

        # Get move for each motor in the channel cut towers towers
        for tower in self._channelcut_towers:
            for motor, position in zip(tower._energy_motors,
                                       tower._get_move_positions(E2)):
                string += "\n{:<15}|{:^15.4f}|{:^15.4f}".format(
                    motor.desc, motor.position, position)

        # Get move for the channel cut diagnostic
        if use_diag:
            position_dcc = self._get_channelcut_diagnostic_position(E2)
            string += "\n{:<15}|{:^15.4f}|{:^15.4f}".format(
                self.parent.dcc.x.desc, self.parent.dcc.x.position,
                position_dcc)

        # Prompt the user for a confirmation or return the string
        if confirm_move is True:
            return self._confirm_move(string)
        else:
            return string

    def _check_towers_and_diagnostics(self, E2, use_diag=_UNSET):
        """
        Checks the staus of the channel cut tower energy motors. Raises the
        basic motor errors if any of the motors are not ready to be moved.

        Parameters
        ----------
        E2 : float
            Desired energy for the channel cut line.

        use_diag : bool, optional
            Check the position of the diagnostic motor.

        Raises
        ------
        LimitError
            Error raised when the inputted position is beyond the soft limits.

        MotorDisabled
            Error raised if the motor is disabled and move is requested.

        MotorFaulted
            Error raised if the motor is disabled and the move is requested.

        MotorStopped
            Error raised If the motor is stopped and a move is requested.

        BadN2Pressure
            Error raised if the pressure in the tower is bad.

        Returns
        -------
        position_dcc : float or None
            Position to move the channel cut diagnostic to.
        """
        use_diag = use_diag if use_diag is not _UNSET else self.use_diag

        # Check each of the channel cut towers
        for tower in self._channelcut_towers:
            tower.check_status(energy=E2)

        # Check the channel cut diagnostic position
        if use_diag:
            position_dcc = self._get_channelcut_diagnostic_position(E2=E2)
            self.parent.dcc.x.check_status(position_dcc)
            return position_dcc

    def _move_towers_and_diagnostics(self, E2, position_dcc, use_diag=_UNSET):
        """
        Moves the channel cut line energy motors and diagnostic to the inputted
        energy and diagnostic position.

        Parameters
        ----------
        E2 : float
            Energy to set the channel cut line to.

        position_dcc : float or None, optional
            Position to move the channel cut diagnostic to.

        use_diag : bool, optional
            Move the daignostic motors to align with the beam.

        Returns
        -------
        status : list
            Nested list of status objects from each tower.
        """
        use_diag = use_diag if use_diag is not _UNSET else self.use_diag
        status = []
        # Move the channel cut towers
        status += [tower.set_energy(E2, wait=False, check_status=False) for
                   tower in self._channelcut_towers]
        # Move the channel cut diagnostics
        if use_diag:
            status += [self.parent.dcc.x.move(position_dcc, wait=False)]
        # Log the energy change
        logger.debug("Setting E2 to {0}.".format(E2))

        return status

    @property
    @nan_if_no_parent
    def position(self):
        """
        Returns the current energy of the channel cut line.

        Returns
        -------
        energy : float
            Energy the channel cut line is set to in eV.
        """
        return self.parent.t2.energy

    def set_position(self, E2=None, print_set=True, verify_move=_UNSET,
                     use_diag=_UNSET):
        """
        Sets the current positions of the motors in the towers to be the
        calculated positions based on the inputted energies or delay.

        Parameters
        ----------
        E2 : float or None, optional
            Energy to set the channel cut line to.

        print_set : bool, optional
            Print a message to the console that the set has been made.

        verify_move : bool, optional
            Prints the current system state and a proposed system state and
            then prompts the user to accept the proposal before changing the
            system.

        use_diag : bool, optional
            Move the daignostic motors to align with the beam.
        """
        verify_move = (verify_move if verify_move is not _UNSET
                       else self._verify_move_option)
        use_diag = use_diag if use_diag is not _UNSET else self.use_diag
        # Prompt the user about the move before making it
        if verify_move and self._verify_move(E2, use_diag=use_diag):
            return

        # Set position of each E2 motor in each channel cut tower
        for tower in self._channelcut_towers:
            for motor, pos in zip(tower._energy_motors,
                                  tower._get_move_positions(E2)):
                motor.set_position(pos, print_set=False)

        # Move the cc diagnostic as well
        position_dcc = self._get_channelcut_diagnostic_position(E2)
        self.parent.dcc.x.set_position(position_dcc, print_set=False)

        # Log the set
        if print_set is True:
            logger.info("Setting positions for E2 to {0}.".format(E2))



"""
Attocube devices
"""
import logging
import os

import numpy as np
from ophyd import Component as Cmp
from ophyd import PositionerBase
from ophyd.signal import EpicsSignal, EpicsSignalRO
from ophyd.status import wait as status_wait
from ophyd.utils import LimitError

from .exceptions import MotorDisabled, MotorError, MotorFaulted
from .snddevice import SndDevice
from .sndmotor import SndMotor
from .utils import absolute_submodule_path, as_list

logger = logging.getLogger(__name__)


class EccController(SndDevice):
    """
    ECC Controller
    """
    _firm_day = Cmp(EpicsSignalRO, ":CALC:FIRMDAY")
    _firm_month = Cmp(EpicsSignalRO, ":CALC:FIRMMONTH")
    _firm_year = Cmp(EpicsSignalRO, ":CALC:FIRMYEAR")
    _flash = Cmp(EpicsSignal, ":RDB:FLASH", write_pv=":CMD:FLASH")

    @property
    def firmware(self):
        """
        Returns the firmware in the same date format as the EDM screen.
        """
        return "{0}/{1}/{2}".format(
            self._firm_day.get(), self._firm_month.get(), self._firm_year.get())

    @property
    def flash(self):
        """
        Saves the current configuration of the controller.
        """
        return self._flash.set(1, timeout=self.set_timeout)


class EccBase(SndMotor, PositionerBase):
    """
    ECC Motor Class
    """
    tab_component_names = True
    tab_whitelist = [
        'check_status',
        # 'check_value',
        'connected',
        'disable',
        'egu',
        'enable',
        'enabled',
        'error',
        'expert_screen',
        'high_limit',
        'limits',
        'low_limit',
        'move',
        'mv',
        'position',
        'reference',
        'referenced',
        'reset',
        'set_limits',
        'status',
        'stop',
    ]

    # position
    user_readback = Cmp(EpicsSignalRO, ":POSITION", auto_monitor=True)
    user_setpoint = Cmp(EpicsSignal, ":CMD:TARGET")

    # limits
    upper_ctrl_limit = Cmp(EpicsSignal, ':CMD:TARGET.HOPR')
    lower_ctrl_limit = Cmp(EpicsSignal, ':CMD:TARGET.LOPR')

    # configuration
    motor_egu = Cmp(EpicsSignalRO, ":UNIT")
    motor_amplitude = Cmp(EpicsSignal, ":CMD:AMPL")
    motor_dc = Cmp(EpicsSignal, ":CMD:DC")
    motor_frequency = Cmp(EpicsSignal, ":CMD:FREQ")

    # motor status
    motor_connected = Cmp(EpicsSignalRO, ":ST_CONNECT")
    motor_enabled = Cmp(EpicsSignalRO, ":ST_ENABLED")
    motor_referenced = Cmp(EpicsSignalRO, ":ST_REFVAL")
    motor_error = Cmp(EpicsSignalRO, ":ST_ERROR")
    motor_is_moving = Cmp(EpicsSignalRO, ":RD_MOVING")
    motor_done_move = Cmp(EpicsSignalRO, ":RD_INRANGE")
    high_limit_switch = Cmp(EpicsSignal, ":ST_EOT_FWD")
    low_limit_switch = Cmp(EpicsSignal, ":ST_EOT_BWD")
    motor_reference_position = Cmp(EpicsSignalRO, ":REF_POSITION")

    # commands
    motor_stop = Cmp(EpicsSignal, ":CMD:STOP")
    motor_reset = Cmp(EpicsSignal, ":CMD:RESET.PROC")
    motor_enable = Cmp(EpicsSignal, ":CMD:ENABLE")

    @property
    def position(self):
        """
        Returns the current position of the motor.

        Returns
        -------
        position : float
            Current position of the motor.
        """
        return self.user_readback.get()

    @property
    def reference(self):
        """
        Returns the reference position of the motor.

        Returns
        -------
        position : float
            Reference position of the motor.
        """
        return self.motor_reference_position.get()

    @property
    def egu(self):
        """
        The engineering units (EGU) for a position

        Returns
        -------
        Units : str
            Engineering units for the position.
        """
        return self.motor_egu.get()

    def _status_print(self, status, msg=None, ret_status=False, print_set=True,
                      timeout=None, wait=True, reraise=False):
        """
        Internal method that optionally returns the status object and optionally
        prints a message about the set. If a message is passed but print_set is
        False then the message is logged at the debug level.

        Parameters
        ----------
        status : StatusObject or list
            The inputted status object.

        msg : str or None, optional
            Message to print if print_set is True.

        ret_status : bool, optional
            Return the status object of the set.

        print_set : bool, optional
            Print a short statement about the set.

        wait : bool, optional
            Wait for the status to complete.

        reraise : bool, optional
            Raise the RuntimeError in the except.

        Returns
        -------
        Status
            Inputted status object.
        """
        try:
            # Wait for the status to complete
            if wait:
                for s in as_list(status):
                    status_wait(s, timeout)

            # Notify the user
            if msg is not None:
                if print_set:
                    logger.info(msg)
                else:
                    logger.debug(msg)
            if ret_status:
                return status

        # The operation failed for some reason
        except RuntimeError:
            error = "Operation completed, but reported an error."
            logger.error(error)
            if reraise:
                raise

    def enable(self, ret_status=False, print_set=True):
        """
        Enables the motor power.

        Parameters
        ----------
        ret_status : bool, optional
            Return the status object of the set.

        print_set : bool, optional
            Print a short statement about the set.

        Returns
        -------
        Status
            The status object for setting the power signal.
        """
        status = self.motor_enable.set(1, timeout=self.set_timeout)
        return self._status_print(status, "Enabled motor '{0}'".format(
            self.desc), ret_status=ret_status, print_set=print_set)

    def disable(self, ret_status=False, print_set=True):
        """
        Disables the motor power.

        Parameters
        ----------
        ret_status : bool, optional
            Return the status object of the set.

        print_set : bool, optional
            Print a short statement about the set.

        Returns
        -------
        Status
            The status object for setting the power signal.
        """
        status = self.motor_enable.set(0, timeout=self.set_timeout)
        return self._status_print(status, "Disabled motor '{0}'".format(
            self.desc), ret_status=ret_status, print_set=print_set)

    @property
    def enabled(self):
        """
        Returns if the motor is curently enabled.

        Returns
        -------
        enabled : bool
            True if the motor is powered, False if not.
        """
        return bool(self.motor_enable.get())

    @property
    def connected(self):
        """
        Returns if the motor is curently connected.

        Returns
        -------
        connected : bool
            True if the motor is connected, False if not.
        """
        return bool(self.motor_connected.get())

    @property
    def referenced(self):
        """
        Returns if the motor is curently referenced.

        Returns
        -------
        referenced : bool
            True if the motor is referenced, False if not.
        """
        return bool(self.motor_referenced.get())

    @property
    def error(self):
        """
        Returns the current error with the motor.

        Returns
        -------
        error : bool
            Error enumeration.
        """
        return bool(self.motor_error.get())

    def reset(self, ret_status=False, print_set=True):
        """
        Sets the current position to be the zero position of the motor.

        Returns
        -------
        status : StatusObject
            Status object for the set.
        """
        status = self.motor_reset.set(1, timeout=self.set_timeout)
        return self._status_print(status, "Reset motor '{0}'".format(
            self.desc), ret_status=ret_status, print_set=print_set)

    def move(self, position, check_status=True, timeout=None, *args, **kwargs):
        """
        Move to a specified position.

        Parameters
        ----------
        position
            Position to move to

        check_status : bool, optional
            Check if the motors are in a valid state to move.

        Returns
        -------
        status : MoveStatus
            Status object for the move.

        Raises
        ------
        TimeoutError
            When motion takes longer than `timeout`

        ValueError
            On invalid positions

        RuntimeError
            If motion fails other than timing out
        """
        # Check the motor status
        if check_status:
            self.check_status(position)
        logger.debug("Moving {0} to {1}".format(self.name, position))
        # Begin the move process
        return self.user_setpoint.set(position, timeout=timeout)

    def mv(self, position, print_move=True, wait=None,
           *args, **kwargs):
        """
        Move to a specified position, optionally waiting for motion to
        complete. mv() is different from move() by catching all the common
        exceptions that this motor can raise and just raises a logger
        warning. Therefore if building higher level functionality, do not
        use this method and use move() instead otherwise none of these
        exceptions will propagate to it.

        Parameters
        ----------
        position
            Position to move to.

        print_move : bool, optional
            Print a short statement about the move.

        Exceptions Caught
        -----------------
        LimitError
            Error raised when the inputted position is beyond the soft limits.

        MotorDisabled
            Error raised if the motor is disabled and move is requested.

        MotorFaulted
            Error raised if the motor is disabled and the move is requested.

        Returns
        -------
        status : MoveStatus
            Status object for the move.
        """
        try:
            status = super().mv(position, *args, **kwargs)

            # Notify the user that a motor has completed or the command is sent
            if print_move:
                logger.info("Move command sent to '{0}'.".format(self.desc))
            # Check if a status object is desired
            return status

        # Catch all the common motor exceptions
        except LimitError:
            logger.warning("Requested move '{0}' is outside the soft limits "
                           "{1} for motor {2}".format(position, self.limits,
                                                      self.desc))
        except MotorDisabled:
            logger.warning("Cannot move - motor {0} is currently disabled. Try "
                           "running 'motor.enable()'.".format(self.desc))
        except MotorFaulted:
            logger.warning("Cannot move - motor {0} is currently faulted. Try "
                           "running 'motor.clear()'.".format(self.desc))

    def check_status(self, position=None):
        """
        Checks the status of the motor to make sure it is ready to move. Checks
        the current position of the motor, and if a position is provided it also
        checks that position.

        Parameters
        ----------
        position : float
            Position to check for validity.

        Raises
        ------
        MotorDisabled
            If the motor is disabled.

        MotorError
            If the motor has an error.
        """
        if not self.enabled:
            err = "Motor '{0}' is currently disabled.".format(self.desc)
            logger.error(err)
            raise MotorDisabled(err)

        if self.error:
            err = "Motor '{0}' currently has an error.".format(self.desc)
            logger.error(err)
            raise MotorError(err)

        # Check if the current position is valid
        self.check_value(self.position)
        # Check if the move position is valid
        if position:
            self.check_value(position)

    def check_value(self, position):
        """
        Checks to make sure the inputted value is valid.

        Parameters
        ----------
        position : float
            Position to check for validity

        Raises
        ------
        ValueError
            If position is None, NaN or Inf
        LimitError
            If the position is outside the soft limits.
        """
        # Check for invalid positions
        if position is None or np.isnan(position) or np.isinf(position):
            raise ValueError("Invalid value inputted: '{0}'".format(position))

        # Check if it is within the soft limits
        if not (self.low_limit <= position <= self.high_limit):
            err_str = (
                "Requested value {0} outside of range: [{1}, {2}]"
                "".format(position, self.low_limit, self.high_limit)
            )
            logger.warn(err_str)
            raise LimitError(err_str)

    def stop(self, success=False, ret_status=False, print_set=True):
        """
        Stops the motor.

        Parameters
        ----------
        ret_status : bool, optional
            Return the status object of the set.

        print_set : bool, optional
            Print a short statement about the set.

        Returns
        -------
        Status : StatusObject
            Status of the set.
        """
        status = self.motor_stop.set(1, wait=False, timeout=self.set_timeout)
        super().stop(success=success)
        return self._status_print(status, "Stopped motor '{0}'".format(
            self.desc), ret_status=ret_status, print_set=print_set)

    def expert_screen(self, print_msg=True):
        """
        Launches the expert screen for the motor.

        Parameters
        ----------
        print_msg : bool, optional
            Prints that the screen is being launched.
        """
        # Get the absolute path to the screen
        path = absolute_submodule_path(
            "hxrsnd/screens/motor_expert_screens.sh")
        if print_msg:
            logger.info("Launching expert screen.")
        os.system("{0} {1} {2} &".format(path, self.prefix, "attocube"))

    def set_limits(self, llm, hlm):
        """
        Sets the limits of the motor. Alias for limits = (llm, hlm).

        Parameters
        ----------
        llm : float
            Desired low limit.

        hlm : float
            Desired low limit.
        """
        self.limits = (llm, hlm)

    @property
    def high_limit(self):
        """
        Returns the upper limit fot the user setpoint.

        Returns
        -------
        high_limit : float
        """
        return self.upper_ctrl_limit.get()

    @high_limit.setter
    def high_limit(self, value):
        """
        Sets the high limit for user setpoint.
        """
        self.upper_ctrl_limit.put(value)

    @property
    def low_limit(self):
        """
        Returns the lower limit fot the user setpoint.

        Returns
        -------
        low_limit : float
        """
        return self.lower_ctrl_limit.get()

    @low_limit.setter
    def low_limit(self, value):
        """
        Sets the high limit for user setpoint.
        """
        self.lower_ctrl_limit.put(value)

    @property
    def limits(self):
        """
        Returns the limits of the motor.

        Returns
        -------
        limits : tuple
        """
        return (self.low_limit, self.high_limit)

    @limits.setter
    def limits(self, value):
        """
        Sets the limits for user setpoint.
        """
        self.low_limit = value[0]
        self.high_limit = value[1]

    def status(self, status="", offset=0, print_status=True, newline=False,
               short=False):
        """
        Returns the status of the device.

        Parameters
        ----------
        status : str, optional
            The string to append the status to.

        offset : int, optional
            Amount to offset each line of the status.

        print_status : bool, optional
            Determines whether the string is printed or returned.

        newline : bool, optional
            Adds a new line to the end of the string.

        Returns
        -------
        status : str
            Status string.
        """
        if short:
            status += "\n{0}{1:<16}|{2:^16.3f}|{3:^16.3f}".format(
                " "*offset, self.desc, self.position, self.reference)
        else:
            status += "{0}{1}\n".format(" "*offset, self.desc)
            status += "{0}PV: {1:>25}\n".format(" "*(offset+2), self.prefix)
            status += "{0}Enabled: {1:>20}\n".format(" "*(offset+2),
                                                     str(self.enabled))
            status += "{0}Faulted: {1:>20}\n".format(" "*(offset+2),
                                                     str(self.error))
            status += "{0}Position: {1:>19}\n".format(" "*(offset+2),
                                                      np.round(self.wm(), 6))
            status += "{0}Limits: {1:>21}\n".format(
                " "*(offset+2), str((int(self.low_limit), int(self.high_limit))))
        if newline:
            status += "\n"
        if print_status is True:
            logger.info(status)
        else:
            return status


class TranslationEcc(EccBase):
    """
    Class for the translation ecc motor
    """
    pass


class GoniometerEcc(EccBase):
    """
    Class for the goniometer ecc motor
    """
    pass


class DiodeEcc(EccBase):
    """
    Class for the diode insertion ecc motor
    """
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import logging
from collections import OrderedDict

import pytest
from ophyd.device import Device

from hxrsnd import aerotech
from hxrsnd.aerotech import AeroBase, MotorDisabled

from .conftest import fake_device, get_classes_in_module

logger = logging.getLogger(__name__)


@pytest.mark.parametrize("dev", get_classes_in_module(aerotech, Device))
def test_aerotech_devices_instantiate_and_run_ophyd_functions(dev):
    motor = fake_device(dev, "TEST:SND:T1")
    assert(isinstance(motor.read(), OrderedDict))
    assert(isinstance(motor.read_configuration(), OrderedDict))


def test_AeroBase_raises_MotorDisabled_if_moved_while_disabled():
    motor = fake_device(AeroBase, "TEST:SND:T1")
    motor.axis_fault.sim_put(0)
    assert not motor.faulted
    motor.disable()
    assert not motor.enabled
    with pytest.raises(MotorDisabled):
        motor.move(10)

# @pytest.mark.parametrize("position", [1])
# def test_AeroBase_callable_moves_the_motor(position):
#     motor = fake_device(AeroBase)
#     motor.axis_fault.sim_put(0)
#     assert not motor.faulted
#     motor.enable()
#     assert motor.enabled
#     motor.limits = (0, 1)
#     assert motor.user_setpoint.get() != position
#     time.sleep(0.5)
#     motor(position, wait=False)
#     time.sleep(0.1)
#     assert motor.user_setpoint.get() == position

# Commented out for now because it causes travis to seg fault sometimes
# def test_AeroBase_raises_MotorFaulted_if_moved_while_faulted():
#     motor = AeroBase("TEST")
#     motor.enable()
#     time.sleep(.1)
#     motor.axis_fault.sim_put(1)
#     with pytest.raises(MotorFaulted):
#         motor.move(10)



import inspect
import logging
import math

import numpy as np
import pandas as pd
import pytest
from bluesky.tests.conftest import RE as fresh_RE  # noqa
from lmfit.models import LorentzianModel
from ophyd.areadetector.base import EpicsSignalWithRBV
from ophyd.device import Component as Cmp
from ophyd.device import Device
from ophyd.signal import Signal
from ophyd.sim import (FakeEpicsSignal, SynAxis, SynSignal, fake_device_cache,
                       make_fake_device)
from pcdsdevices.areadetector.detectors import PCDSAreaDetector

from ..sndmotor import CalibMotor

logger = logging.getLogger(__name__)

# Define the requires epics
try:
    import epics
    pv = epics.PV("XCS:USR:MMS:01")
    try:
        val = pv.get()
    except Exception:
        val = None
except Exception:
    val = None


epics_subnet = val is not None
requires_epics = pytest.mark.skipif(not epics_subnet,
                                    reason="Could not connect to sample PV")


# Enable the logging level to be set from the command line
def pytest_addoption(parser):
    parser.addoption("--log", action="store", default="INFO",
                     help="Set the level of the log")
    parser.addoption("--logfile", action="store", default=None,
                     help="Write the log output to specified file path")


class Diode(SynSignal):
    """
    Simulated Diode

    Evaluate a point on a Lorentz function based on the value of a motor

    By default, the amplitude and sigma values will create a max signal of 1.0,
    representing a normalized diode signal

    Parameters
    ----------
    name : str

    motor : obj

    motor_field : str
        Name of field to use as independent variable

    center : float
        Center position of Lorentz

    sigma : float, optional
        Width of distribution

    amplitude : float, optional
        Height of distribution

    noise_multiplier : float, optional
        Multipler for uniform noise of the diode. If left as None, no noise will
        be applied
    """
    def __init__(self, name, motor, motor_field, center,
                 sigma=1, amplitude=math.pi,
                 noise_multiplier=None, **kwargs):
        # Eliminate noise if not requested
        noise = noise_multiplier or 0.
        lorentz = LorentzianModel()

        def func():
            # Evaluate position in distribution
            m = motor.read()[motor_field]['value']
            v = lorentz.eval(x=np.array(m), amplitude=amplitude, sigma=sigma,
                             center=center)
            # Add uniform noise
            v += np.random.uniform(-1, 1) * noise
            return v

        # Instantiate Reader
        super().__init__(name=name, func=func, **kwargs)


class SynCentroid(SynSignal):
    """
    Synthetic centroid signal.
    """
    def __init__(self, motors, weights, noise_multiplier=None, name=None,
                 *args, **kwargs):
        # Eliminate noise if not requested
        self.motors = motors
        self.weights = weights
        self.noise = noise_multiplier or 0.

        def func():
            # Evaluate the positions of each motor
            pos = [m.position for m in self.motors] or [0, 0]
            # Get the centroid position
            cent = np.dot(pos, self.weights)
            # Add uniform noise
            cent += int(np.round(np.random.uniform(-1, 1) * self.noise))
            return cent

        # Instantiate the synsignal
        super().__init__(name=name, func=func, **kwargs)


class SynCamera(Device):
    """
    Simulated camera that has centroids as components.
    """
    centroid_x = Cmp(SynCentroid, motors=[], weights=[1, 0.25])
    centroid_y = Cmp(SynCentroid, motors=[], weights=[1, -0.25])

    def __init__(self, motor1, motor2, delay, name=None, *args, **kwargs):
        # Create the base class
        super().__init__("SYN:CAMERA", name=name, *args, **kwargs)

        # Define the centroid components using the inputted motors
        self.centroid_x.motors = [motor1, delay]
        self.centroid_y.motors = [motor2, delay]

        # Add them to _signals
        self._signals['centroid_x'] = self.centroid_x
        self._signals['centroid_y'] = self.centroid_y
        # Add them to the read_attrs
        self.read_attrs = ["centroid_x", "centroid_y"]

    def trigger(self):
        return self.centroid_x.trigger() & self.centroid_y.trigger()


class CalibTest(CalibMotor):
    motor = Cmp(SynAxis, name="test_axis")

    def __init__(self, *args, name="calib", m1=None, m2=None, **kwargs):
        super().__init__(*args, name="calib", **kwargs)
        self.calib_detector = SynCamera(m1, m2, self.motor, name="camera")
        self.calib_motors = [m1, m2]
        self.motor_fields = [self.motor.name]
        self.detector_fields = ['centroid_x', 'centroid_y']
        self.set = self.move
        for m in [self.motor] + self.calib_motors:
            m.move = m.set

    @property
    def position(self):
        return self.motor.position

    def move(self, position, *args, **kwargs):
        # Perform the calibration move
        status = self.motor.set(position, *args, *kwargs)
        if self.has_calib and self.use_calib:
            status = status & self._calib_compensate(position)
        return status


# Simulated Crystal motor that goes where you tell it
crystal = SynAxis(name='angle')
m1 = SynAxis(name="m1")
m2 = SynAxis(name="m2")
delay = SynAxis(name="delay")


# Create a fixture to automatically instantiate logging setup
@pytest.fixture(scope='session', autouse=True)
def set_level(pytestconfig):
    # Read user input logging level
    log_level = getattr(logging, pytestconfig.getoption('--log'), None)

    # Report invalid logging level
    if not isinstance(log_level, int):
        raise ValueError("Invalid log level : {}".format(log_level))

    # Create basic configuration
    logging.basicConfig(level=log_level,
                        filename=pytestconfig.getoption('--logfile'))


@pytest.fixture(scope='function')
def get_calib_motor(request):
    m1 = SynAxis(name="m1")
    m2 = SynAxis(name="m2")
    return CalibTest("test", m1=m1, m2=m2)


def get_classes_in_module(module, subcls=None, blacklist=None):
    classes = []
    blacklist = blacklist or list()
    all_classes = [cls for _, cls in inspect.getmembers(module)
                   if cls not in blacklist]
    for cls in all_classes:
        try:
            if cls.__module__ == module.__name__:
                if subcls is not None:
                    try:
                        if not issubclass(cls, subcls):
                            continue
                    except TypeError:
                        continue
                classes.append(cls)
        except AttributeError:
            pass
    return classes


# Create a fake epics device
def fake_device(device, name="TEST"):
    device = make_fake_device(device)
    return device(name, name=name)


def fake_detector(detector, name="TEST"):
    """Set the plugin_type signal to be _plugin_type for all plugins."""
    def change_all_plugin_types(comp):
        if hasattr(comp, "component_names"):
            if hasattr(comp, "_plugin_type"):
                comp.plugin_type = Cmp(Signal, value=comp._plugin_type)
            else:
                for name in comp.component_names:
                    sub_comp = getattr(comp, name)
                    if type(sub_comp) is Cmp:
                        sub_comp = change_all_plugin_types(sub_comp.cls)
        return comp
    detector = change_all_plugin_types(detector)
    detector = make_fake_device(detector)
    return detector(name, name=name)


# Hotfix area detector plugins for tests
for comp in (PCDSAreaDetector.image1, PCDSAreaDetector.stats2):
    plugin_class = comp.cls
    plugin_class.plugin_type = Cmp(Signal, value=plugin_class._plugin_type)

# Hotfix make_fake_device for ophyd=1.2.0
fake_device_cache[EpicsSignalWithRBV] = FakeEpicsSignal

test_df_scan = pd.DataFrame(
    [[-1, 0, 0, -0.25, 0.25],
     [-0.5, 0, 0, -0.125, 0.125],
     [0, 0, 0, 0, 0],
     [0.5, 0, 0, 0.125, -0.125],
     [1, 0, 0, 0.25, -0.25]
     ],
    index=np.linspace(-1, 1, 5),
    columns=["delay", "m1_pre", "m2_pre", "camera_centroid_x",
             "camera_centroid_y"]
)



import logging

import numpy as np
from bluesky.preprocessors import run_wrapper
from ophyd.sim import SynAxis

from ..plans.alignment import maximize_lorentz, rocking_curve
from .conftest import Diode

logger = logging.getLogger(__name__)

crystal = SynAxis(name='angle')


def test_lorentz_maximize(fresh_RE):
    # Simulated diode readout
    diode = Diode('intensity', crystal, 'angle', 10.0, noise_multiplier=None)
    # Create plan to maximize the signal
    plan = run_wrapper(maximize_lorentz(diode, crystal, 'intensity',
                                        step_size=0.2, bounds=(9., 11.),
                                        position_field='angle',
                                        initial_guess={'center': 8.}))
    # Run the plan
    fresh_RE(plan)

    # Trigger an update
    diode.trigger()
    # Check that we were within 10%
    assert np.isclose(diode.read()['intensity']['value'], 1.0, 0.1)


def test_rocking_curve(fresh_RE):
    # Simulated diode readout
    diode = Diode('intensity', crystal, 'angle', 10.0, noise_multiplier=None)
    # Create plan to maximize the signalplan
    plan = run_wrapper(rocking_curve(diode, crystal, 'intensity',
                                     coarse_step=0.1, fine_step=0.05,
                                     bounds=(5., 15.), fine_space=2.5,
                                     position_field='angle',
                                     initial_guess={'center': 8.}))
    # Run the plan
    fresh_RE(plan)

    # Trigger an update
    diode.trigger()
    # Check that we were within 10%
    assert np.isclose(diode.read()['intensity']['value'], 1.0, 0.1)
